from tcutility.results import Result
from tcutility import constants, slurm
import os
from scm import plams
import numpy as np


j = os.path.join


def get_calc_files(calc_dir: str) -> Result:
    """Function that returns files relevant to ORCA calculations stored in ``calc_dir``.

    Args:
        calc_dir: path pointing to the desired calculation

    Returns:
        Result object containing filenames and paths
    """
    # collect all files in the current directory and subdirectories
    files = []
    for root, _, files_ in os.walk(calc_dir):
        files.extend([j(root, file) for file in files_])

    # we now go through all the files and check their nature
    ret = Result()
    ret.root = os.path.abspath(calc_dir)
    for file in files:
        with open(file, errors='ignore') as f:
            lines = f.readlines()
            # detect the output file
            if any(["* O   R   C   A *" in line for line in lines]):
                ret.out = os.path.abspath(file)
                continue    

            # detect the input file
            # there should be lines starting with ! and also the system line, starting with * xyz, * xyzfile, * gzmtfile or * int
            if any([line.startswith('!') for line in lines]) and any([(len(line.split()) > 2 and line.split()[0] == '*' and line.split()[1] in ['xyz', 'xyzfile', 'gzmtfile', 'int']) for line in lines]):
                ret.inp = os.path.abspath(file)
                continue

    return ret


def get_version(info: Result) -> Result:
    """Function to get the ORCA version used in the calculation.

    Args:
        info: Result object containing ORCA calculation settings.

    Returns:
        :Result object containing results about the ORCA version:

            - **full (str)** – the full version string as written by ORCA.
            - **major (str)** – major ORCA version.
            - **minor (str)** – minor ORCA version.
            - **micro (str)** – micro ORCA version.
    """
    ret = Result()
    with open(info.files.out) as out:
        for line in out.readlines():
            line = line.strip()
            if "Program Version" not in line:
                continue
            version = line.split()[2]
            ret.full = version
            ret.major = version.split(".")[0]
            ret.minor = version.split(".")[1]
            ret.micro = version.split(".")[2]
            return ret


def get_input(info: Result) -> Result:
    """Function that parses the input file for this ORCA calculation.

    Args:
        info: Result object containing ORCA calculation settings.

    Returns:
        :Result object containing information about the calculation input:

            - **main (list[str])** - the main inputs for the calculation. These are the lines that start with a "!".
            - **sections (Result)** - extra settings added to the calculation. These are the lines that start with a "%" and optionally end with "END" clause.
            - **system (Result)** - settings related to the molecular system. This includes charge, multiplicity and the coordinates.
            - **task (str)** - the task that was performed by the calculation, e.g. "SinglePoint", "TransitionStateSearch".
    """
    ret = Result()

    # we read the input file first
    if 'inp' in info.files:
        with open(info.files.inp) as inp:
            lines = inp.readlines()
    # if we don't have it we read the output file, which contains the input as a block
    else:
        with open(info.files.out) as out:
            start_reading = False
            lines = []
            for line in out.readlines():
                line = line.strip()
                if start_reading:
                    lines.append(line)

                if "INPUT FILE" in line:
                    start_reading = True
                    continue

                if "****END OF INPUT****" in line:
                    break

        lines = [line.split(">")[1] for line in lines[2:-1] if line.split(">")[1].strip()]

    ret.main = []
    curr_section = None
    read_system = False
    system_lines = []
    coordinates = None
    for line in lines:
        line = line.strip()

        if line.startswith("!"):
            ret.main.extend(line.strip("!").split())

        if curr_section:
            if line.lower() == "end":
                curr_section = None
                continue

            var, val = line.split()
            ret.sections[curr_section][var] = val

        if line.startswith("%"):
            curr_section = line.split()[0][1:]
            if len(line.split()) == 2:
                ret.sections[curr_section] = line.split()[1]
                curr_section = None

        if read_system:
            if line == "*":
                read_system = False
                continue

            system_lines.append(line)

        if line.startswith("*"):
            read_system = True
            _, coordinates, charge, multiplicity = line.split()[:4]
            if coordinates == "xyz":
                ret.system.coordinate_system = "cartesian"
            elif coordinates == "int":
                ret.system.coordinate_system = "internal"
            elif coordinates == "xyzfile":
                ret.system.coordinate_system = "cartesian"
                read_system = False
            ret.system.charge = charge
            ret.system.multiplicity = multiplicity
            continue

    if coordinates in ["xyz", "int"]:
        ret.system.molecule = plams.Molecule()
        for line in system_lines:
            line = line.replace(':', '')
            ret.system.molecule.add_atom(plams.Atom(symbol=line.split()[0], coords=[float(x) for x in line.split()[1:4]]))

    ret.task = "SinglePoint"
    if "optts" in [x.lower() for x in ret.main]:
        ret.task = "TransitionStateSearch"
    elif "opt" in [x.lower() for x in ret.main]:
        ret.task = "GeometryOptimization"

    return ret


def get_level_of_theory(info: Result) -> Result:
    """Function to get the level-of-theory from an input-file.

    Args:
        info: Result object containing ORCA calculation settings.

    Returns:
        :Result object containing information about the level-of-theory:

            - **summary (str)** - a summary string that gives the level-of-theory in a human-readable format.
            - **basis.type (str)** - the size/type of the basis-set.
            - **UsedQROs (bool)** - whether QROs were used.
    """
    sett = info.input
    ret = Result()
    main = [x.lower() for x in sett.main]
    ret.method = 'HF'
    for method in ["MP2", "CCSD", "CCSD(T)", "CCSDT"]:
        if method.lower() in main:
            ret.method = method
            break

    ret.basis.type = 'def2-SVP'
    for bs in ["cc-pVDZ", "cc-pVTZ", "cc-pVQZ", "cc-pV5Z", "aug-cc-pVDZ", "aug-cc-pVTZ", "aug-cc-pVQZ", "aug-cc-pV5Z"]:
        if bs.lower() in main:
            ret.basis.type = bs

    used_qros = sett.sections.mdci.UseQROs and sett.sections.mdci.UseQROs.lower() == "true"
    ret.summary = f'{"QRO-" if used_qros else ""}{method}/{ret.basis.type}'

    return ret


def get_calc_settings(info: Result) -> Result:
    """Function to read calculation settings for an ORCA calculation.

    Args:
        info: Result object containing ORCA calculation settings.

    Returns:
        :Result object containing properties from the ORCA calculation:

            - **task (str)** – the task that was set for the calculation.
            - **unrestricted (bool)** – whether or not the wavefunction is treated in an unrestricted manner.
            - **used_qros (bool)** - whether the reference wavefunction is transformed to a QRO wavefunction.
            - **frequencies (bool)** - whether vibrational frequencies were calculated.
            - **charge (int)** - the charge of the molecular system.
            - **spin_polarization (int)** - the spin-polarization of the molecular system.
            - **multiplicity (int)** - the multiplicity of the molecular system.
    """

    assert info.engine == "orca", f"This function reads ORCA data, not {info.engine} data"

    ret = Result()

    # set the calculation task at a higher level
    ret.task = info.input.task

    main = [x.lower() for x in info.input.main]
    # determine if the wavefunction are unrestricted or not
    ret.unrestricted = any(tag in main for tag in ["uhf", "uno"])
    ret.used_qros = info.input.sections.mdci.UseQROs and info.input.sections.mdci.UseQROs.lower() == "true"
    ret.frequencies = "freq" in main or 'numfreq' in main
    ret.charge = int(info.input.system.charge)
    ret.spin_polarization = int(info.input.system.multiplicity) - 1
    ret.multiplicity = int(info.input.system.multiplicity)
    return ret


def get_calculation_status(info: Result) -> Result:
    """Function that returns the status of the ORCA calculation described in reader. In case of non-succes it will also give possible reasons for the errors/warnings.

    Args:
        info: Result object containing ORCA calculation information.

    Returns:
        :Result object containing information about the calculation status:

            - **fatal (bool)** – True if calculation cannot be analysed correctly, False otherwise
            - **reasons (list[str])** – list of reasons to explain the status, they can be errors, warnings, etc.
            - **name (str)** – calculation status written as a string, one of ("SUCCESS", "RUNNING", "UNKNOWN", "SUCCESS(W)", "FAILED")
            - **code (str)** – calculation status written as a single character, one of ("S", "R", "U", "W" "F")
    """
    ret = Result()
    ret.fatal = True
    ret.name = None
    ret.code = None
    ret.reasons = []

    # if we do not have an output file the calculation failed
    if "out" not in info.files.out:
        ret.reasons.append("Calculation status unknown")
        ret.name = "UNKNOWN"
        ret.code = "U"
        return ret

    # try to read if the calculation succeeded
    with open(info.files.out) as out:
        lines = out.readlines()
        if any(["ORCA TERMINATED NORMALLY" in line for line in lines]):
            ret.fatal = False
            ret.name = "SUCCESS"
            ret.code = "S"
            return ret

    # if it didnt we default to failed
    ret.name = "FAILED"
    ret.code = "F"

    # otherwise we check if the job is being managed by slurm
    if not slurm.workdir_info(os.path.abspath(info.files.root)):
        return ret

    # get the statuscode from the workdir
    state = slurm.workdir_info(os.path.abspath(info.files.root)).statuscode
    state_name = {
        'CG': 'COMPLETING',
        'CF': 'CONFIGURING',
        'PD': 'PENDING',
        'R': 'RUNNING'
    }.get(state, 'UNKNOWN')

    ret.fatal = False
    ret.name = state_name
    ret.code = state
    ret.reasons = []

    return ret


def get_molecules(info: Result) -> Result:
    """Function that returns information about the molecules for this calculation.

    Args:
        info: Result object containing ORCA calculation information.

    Returns:
        :Result object containing properties from the ORCA calculation:

            - **input (plams.Molecule)** - the input molecule for this calculation.
            - **output (plams.Molecule)** - the output molecule for this calculation, for example the optimized structure after a geometry optimization.
            - **number_of_atoms (int)** - the number of atoms in the molecular system.
    """
    ret = Result()
    ret.input = info.input.system.molecule
    ret.number_of_atoms = len(ret.input.atoms)

    with open(info.files.out) as out:
        lines = out.readlines()
        lines = [line.strip() for line in lines]

    start_reading = False
    look_for_coords = False
    coords = []
    for line in lines:
        if start_reading:
            if len(line) == 0:
                start_reading = False
                break

            coords.append(line)

        if "THE OPTIMIZATION HAS CONVERGED" in line:
            look_for_coords = True

        if look_for_coords and "CARTESIAN COORDINATES (ANGSTROEM)" in line:
            look_for_coords = False
            start_reading = True

    ret.output = plams.Molecule()
    for coord in coords[1:]:
        sym, x, y, z = coord.split()
        ret.output.add_atom(plams.Atom(symbol=sym, coords=[float(x), float(y), float(z)]))

    if len(ret.output.atoms) == 0:
        ret.output = ret.input.copy()
    return ret


def get_info(calc_dir: str) -> Result:
    """Function to read useful info about the calculation in ``calc_dir``. Returned information will depend on the type of file that is provided.

    Args:
        calc_dir: path pointing to the desired calculation.

    Returns:
        :Result object containing results about the calculation and AMS:

            - **version (Result)** – information about the AMS version used, see :func:`get_version`.
            - **files (Result)** - paths to files that are important for this calculation.
            - **input (Result)** - information about the input of this calculation, see :func:`get_input`.
            - **level (Result)** - information about the level of theory used for this calculation, see :func:`get_level_of_theory`.
            - **engine (str)** – the engine that was used to perform the calculation, for example 'adf', 'dftb', ...
            - **status (Result)** – information about calculation status, see :func:`get_calculation_status`.
            - **molecule (Result)** – information about the input and output molecules and the molecular system in general, see :func:`get_molecules`.
    """
    ret = Result()

    ret.engine = "orca"
    ret.files = get_calc_files(calc_dir)

    # store the input of the calculation
    ret.input = get_input(ret)

    ret.level = get_level_of_theory(ret)

    # store information about the version of AMS
    ret.version = get_version(ret)

    # store the calculation status
    ret.status = get_calculation_status(ret)

    # read molecules
    ret.molecule = get_molecules(ret)

    return ret


def get_vibrations(lines):
    """Function to read vibrational data of an ORCA calculation.

    Args:
        lines: Lines in the output file of the ORCA calculation.

    Returns:
        :Result object containing vibrational properties from the ORCA calculation:

            - **number_of_modes (int)** – number of vibrational modes for this molecule, 3N-5 for non-linear molecules and 3N-6 for linear molecules, where N is the number of atoms.
            - **number_of_imaginary_modes (int)** – number of imaginary vibrational modes for this molecule.
            - **frequencies (list[float])** – vibrational frequencies associated with the vibrational modes, sorted from low to high (|cm-1|).
            - **intensities (list[float])** – vibrational intensities associated with the vibrational modes (|km/mol|). In ORCA, intensities of imaginary modes are set to 0.
            - **modes (list[float])** – list of vibrational modes sorted from low frequency to high frequency.
            - **character (str)** – the PES character of the molecular system. Either "minimum", "transitionstate" or "saddlepoint(n_imag)", for 0, 1, n_imag number of imaginary frequencies.
    """
    ret = Result()
    start_reading = False
    start_reading_idx = 0
    freq_lines = []
    for i, line in enumerate(lines):
        if start_reading:
            if len(line) == 0 and (i - start_reading_idx) > 4:
                break
            if ":" in line:
                freq_lines.append(line)

        if "VIBRATIONAL FREQUENCIES" in line:
            start_reading = True
            start_reading_idx = i

    ret.number_of_modes = len(freq_lines)
    frequencies = [float(line.split()[1]) for line in freq_lines]
    nrotranslational = sum([freq == 0 for freq in frequencies])
    ret.frequencies = frequencies[nrotranslational:]
    ret.number_of_imaginary_modes = len([freq for freq in ret.frequencies if freq < 0])
    ret.character = "minimum" if ret.number_of_imaginary_modes == 0 else "transitionstate" if ret.number_of_imaginary_modes == 1 else f"saddlepoint({ret.number_of_imaginary_modes})"

    start_reading = False
    mode_lines = []
    for i, line in enumerate(lines):
        if "NORMAL MODES" in line:
            start_reading = True
            continue

        if "IR SPECTRUM" in line:
            start_reading = False
            break

        if start_reading:
            mode_lines.append(line)

    mode_lines = mode_lines[6:-3]
    mode_lines = [[float(x) for x in line.split()[1:]] for i, line in enumerate(mode_lines) if i % (ret.number_of_modes + 1) != 0]

    nblocks = len(mode_lines) // ret.number_of_modes
    blocks = []
    for block in range(nblocks):
        blocks.append(np.array(mode_lines[block * ret.number_of_modes : (block + 1) * ret.number_of_modes]))
    ret.modes = np.hstack(blocks).T.tolist()[nrotranslational:]

    start_reading = False
    int_lines = []
    for i, line in enumerate(lines):
        if "IR SPECTRUM" in line:
            start_reading = True
            continue

        if "The epsilon (eps) is given for a Dirac delta lineshape." in line:
            start_reading = False
            break

        if start_reading:
            int_lines.append(line)

    ints = [float(line.split()[3]) for line in int_lines[5:-1]]
    ret.intensities = [0] * ret.number_of_imaginary_modes + ints
    return ret


def get_properties(info: Result) -> Result:
    """Function to get properties from an ORCA calculation.

    Args:
        info: Result object containing ORCA properties.

    Returns:
        :Result object containing properties from the ORCA calculation:

            - **energy.bond (float)** – total bonding energy (|kcal/mol|).
            - **energy.enthalpy (float)** – enthalpy (|kcal/mol|). Only obtained if vibrational modes were calculated.
            - **energy.entropy (float)** – entropy (|kcal/mol|). Only obtained if vibrational modes were calculated.
            - **energy.gibbs (float)** – Gibb's free energy (|kcal/mol|). Only obtained if vibrational modes were calculated.
            - **energy.[method] (float)** - total energy (|kcal/mol|) at a certain level (HF, MP2, CCSD, ...). This is the sum of energy.HF and energy.[method]_corr.
            - **energy.[method]_corr (float)** - electron correlation energy (|kcal/mol|) at a certain level (HF, MP2, CCSD, ...).
            - **vibrations.number_of_modes (int)** – number of vibrational modes for this molecule, 3N-5 for non-linear molecules and 3N-6 for linear molecules, where N is the number of atoms.
            - **vibrations.number_of_imaginary_modes (int)** – number of imaginary vibrational modes for this molecule.
            - **vibrations.frequencies (list[float])** – vibrational frequencies associated with the vibrational modes, sorted from low to high (|cm-1|).
            - **vibrations.intensities (list[float])** – vibrational intensities associated with the vibrational modes (|km/mol|). In ORCA, intensities of imaginary modes are set to 0.
            - **vibrations.modes (list[float])** – list of vibrational modes sorted from low frequency to high frequency.
            - **vibrations.character (str)** – the PES character of the molecular system. Either "minimum", "transitionstate" or "saddlepoint(n_imag)", for 0, 1, n_imag number of imaginary frequencies.
            - **t1** - T1 diagnostic for the highest level of correlation chosen. Used to check the validity of the reference wavefunction.
            - **s2** - expectation value of the :math:`S^2` operator.
            - **s2_expected** - ideal expectation value of the :math:`S^2` operator.
            - **spin_contamination** - the amount of spin-contamination observed in this calculation. It is equal to (s2 - s2_expected) / (s2_expected). Ideally this value should be below 0.1.
    """
    ret = Result()

    with open(info.files.out) as out:
        lines = [line.strip() for line in out.readlines()]

    if info.orca.frequencies:
        ret.vibrations = get_vibrations(lines)

    # read some important info about the calculation
    for line in lines:
        if "FINAL SINGLE POINT ENERGY" in line:
            ret.energy.bond = float(line.split()[4]) * constants.HA2KCALMOL
            continue

        if "E(0)" in line:
            ret.energy.HF = float(line.split()[-1]) * constants.HA2KCALMOL
            continue

        if "Final correlation energy" in line:
            ret.energy.corr = float(line.split()[-1]) * constants.HA2KCALMOL
            continue

        if "E(MP2)" in line:
            ret.energy.MP2 = float(line.split()[-1]) * constants.HA2KCALMOL + ret.energy.HF
            ret.energy.MP2_corr = float(line.split()[-1]) * constants.HA2KCALMOL
            continue

        if "E(CCSD) " in line:
            ret.energy.CCSD = float(line.split()[-1]) * constants.HA2KCALMOL
            ret.energy.CCSD_corr = float(line.split()[-1]) * constants.HA2KCALMOL - ret.energy.HF
            continue

        if "E(CCSD(T))" in line:
            ret.energy.CCSD_T = float(line.split()[-1]) * constants.HA2KCALMOL
            ret.energy.CCSD_T_corr = float(line.split()[-1]) * constants.HA2KCALMOL - ret.energy.HF
            continue

        if "Final Gibbs free energy" in line:
            ret.energy.gibbs = float(line.split()[-2]) * constants.HA2KCALMOL
            continue

        if "Total enthalpy" in line:
            ret.energy.enthalpy = float(line.split()[-2]) * constants.HA2KCALMOL
            continue

        if "Final entropy term" in line:
            ret.energy.entropy = float(line.split()[-2])
            continue

        if "T1 diagnostic" in line:
            ret.t1 = float(line.split()[3])
            continue

        if "Expectation value of <S**2>" in line:
            ret.s2 = float(line.split()[-1])
            continue

        if "Ideal value" in line:
            ret.s2_expected = float(line.split()[-1])
            continue

    if ret.s2 and ret.s2_expected:
        ret.spin_contamination = (ret.s2 - ret.s2_expected) / ret.s2_expected

    return ret




'''Module containing the TCutility.results.result.Result class.'''
import dictfunc
import sys


class Result(dict):
    '''Class used for storing results from AMS calculations. The class is functionally a dictionary, but allows dot notation to access variables in the dictionary. 
    The class works case-insensitively, but will retain the case of the key when it was first set.'''
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        # if we are casting an object to a Result object
        # we will copy all data to this one and all dictionaries will be turned into Result object
        # turn every value into a Result object if possible
        for key, value in self.items():
            if isinstance(value, dict):
                self[key] = Result(value)
            else:
                self[key] = value

    def __call__(self):
        '''Calling of a dictionary subclass should not be possible, instead we raise an error with information about the key and method that were attempted to be called.'''
        head, method = '.'.join(self.get_parent_tree().split('.')[:-1]), self.get_parent_tree().split('.')[-1]
        raise AttributeError(f'Tried to call method "{method}" from {head}, but {head} is empty')

    def __str__(self):
        '''Override str method to prevent printing of hidden keys. You can still print them if you call repr instead of str.'''
        return '{' + ', '.join([f'{key}: {str(val)}' for key, val in self.items()]) + '}'

    def items(self):
        '''We override the items method from dict in order to skip certain keys. We want to hide keys starting and ending
        with dunders, as they should not be exposed to the user.
        '''
        return [(key, self[key]) for key in self.keys()]

    def keys(self):
        original_keys = super().keys()
        return [key for key in original_keys if not (key.startswith('__') and key.endswith('__'))]

    def multi_keys(self):
        '''
        Return multi_keys for this Result object. These are unnested keys that can be used if you want a flattened Result object.
        '''
        def dict_to_list(a):
            '''
            Return a nested dictionary as a list of keys and values.
            '''
            lst = []
            if not isinstance(a, dict):
                return [[a]]
            for k, v in a.items():
                if isinstance(v, dict):
                    if v == {}:
                        lst.append([k, {}])
                    else:
                        [lst.append([k, *x]) for x in dict_to_list(v)]
                else:
                    lst.append([k, v])
            return lst

        # cast this object to a list of keys and values
        mks = dict_to_list(self)
        # write the multi-keys separated with dots
        mks = ['.'.join(mk[:-1]) for mk in mks]
        return mks

    def __getitem__(self, key):
        if key.startswith('__') and key.endswith('__'):
            return None
        self.__set_empty(key)
        val = super().__getitem__(self.__get_case(key))
        return val

    def __getattr__(self, key):
        return self.__getitem__(key)

    def __setitem__(self, key, val):
        # we set the item, but if it is a dict we convert the dict to a Result first
        if isinstance(val, dict):
            val = Result(val)
        super().__setitem__(self.__get_case(key), val)

    def __setattr__(self, key, val):
        self.__setitem__(key, val)

    def __contains__(self, key):
        # Custom method to check if the key is defined in this object and is also non-empty, case-insensitive.
        return key.lower() in [key_.lower() for key_ in self.keys()] and self[key]

    def __hash__(self):
        '''Hashing of a dictionary subclass should not be possible, instead we should raise an error to let the user know
        that they made a mistake. Also give information of which key was being read.
        '''
        raise KeyError(f'Tried to hash {self.get_parent_tree()}, but it is empty')

    def __reduce__(self):
        return 'TCutility.results.result.Result'

    def __bool__(self):
        '''Make sure that keys starting and ending in "__" are skipped'''
        return len([key for key in self.keys() if not (key.startswith('__') and key.endswith('__'))]) > 0

    def __sizeof__(self):
        '''
        Magic method used by `sys.getsizeof <https://docs.python.org/3/library/sys.html#sys.getsizeof>`_ to determine the memory footprint of this object.
        '''
        s = super().__sizeof__()
        for key in self.multi_keys():
            s += sys.getsizeof(self.get_multi_key(key))
        return s

    def getsizeof(self):
        '''
        Return the size of this object in bytes.
        '''
        return self.__sizeof__()

    def get_parent_tree(self):
        '''Method to get the path from this object to the parent object. The result is presented in a formatted string'''
        # every parent except the top-most parent has defined a __parent__ attribute
        if '__parent__' not in self:
            return 'Head'
        # iteratively build the tree using the __name__ attribute.
        parent_names = self.__parent__.get_parent_tree()
        parent_names += '.' + self.__name__
        return parent_names

    def __set_empty(self, key):
        # This function checks if the key has been set. 
        # If it has not, we create a new Result object and set it at the desired key
        if self.__get_case(key) not in self.keys():
            val = Result()
            # we also keep track of the parent of this object and also the name it was assigned to for later bookkeeping
            val.__parent__ = self
            val.__name__ = key
            self.__setitem__(key, val)

    def __get_case(self, key):
        # Get the case of the key as it has been set in this object.
        # The first time a key-value pair has been assigned the case of the key will be set.
        for key_ in self:
            if key_.lower() == key.lower():
                return key_
        return key

    def prune(self):
        '''Remove empty paths of this object.
        '''
        items = list(self.items())
        for key, value in items:
            try:
                value.prune()
            except AttributeError:
                pass

            if not value:
                del self[key]

    def get_multi_key(self, key: str):
        '''
        Method that returns the value of a "multikey". The multikey is multiple keys joined by dots. 
        E.g. res.properties.energy.bond can be gotten by calling res.get_multi_key("properties.energy.bond")
        '''
        data = self
        for keypart in key.split('.'):
            data = data[keypart]
        return data

    def as_plams_settings(self):
        '''
        Returns this Result object as a plams.Settings object.
        '''
        from scm import plams
        import dictfunc

        clean_dict = dictfunc.list_to_dict(dictfunc.dict_to_list(self))
        return plams.Settings(clean_dict)

    def copy(self):
        import copy

        # cast this object to a list of keys and values
        lsts = dictfunc.dict_to_list(self)
        # copy everthing in the lists
        lsts = [[copy.copy(x) for x in lst] for lst in lsts]
        # and return a new result object
        return Result(dictfunc.list_to_dict(lsts))


if __name__ == '__main__':
    ret = Result()
    # print(ret.adf)
    # print(dict(ret.adf))
    # print(bool(ret.adf))
    ret.adf.x = {'a': 1, 'b': 2}
    # ret.adf.system.atoms = []
    # ret.adf.system.atoms.append('test 1 2 3')

    # test_dict[ret.adf.y] = 20
    # ret.adf.y.join()
    # {ret.test: 123}
    # ret.__name__ = 'testname'
    # print(ret.__name__)
    print(repr(ret))



from tcutility import pathfunc, results
import json


degeneracies = {
	'S': 1,
	'P': 3,
	'D': 5,
	'F': 7,
	'G': 9,
}

ret_data = results.Result()
for d, info in pathfunc.match('/Applications/AMS2023.104.app/Contents/Resources/amshome/atomicdata/ADF/ZORA', '{BS}/{element}').items():
	if '.' in info.element:
		continue

	if info.element == 'Ni_d9':
		continue

	if info.element == '00README':
		continue
	
	with open(d) as bsfile:
		lines = bsfile.readlines()

	lines = [line.strip() for line in lines if line.strip()]

	orbs = []
	start_reading = False
	for line in lines:
		if line.lower() == 'basis':
			start_reading = True
			continue

		if start_reading:
			if line.lower() == 'end':
				break
			orbs.append(line.split()[0])

	norbs = sum([degeneracies[orb[-1].upper()] for orb in orbs])
	ret_data[info.bs][info.element] = norbs

with open('_atom_data_info/norbs.json', 'w+') as out:
	out.write(json.dumps(ret_data, indent=4))



import os
import platform
import subprocess as sp
import time

from tcutility import cache, log, results


@cache.cache
def has_slurm() -> bool:
    """
    Function to check if the current platform uses slurm.

    Returns:
        Whether slurm is available on this platform.
    """
    try:
        # Determine the appropriate command based on the OS
        command = ["which", "sbatch"] if platform.system() != "Windows" else ["where", "sbatch"]

        # we do not want this function to print anything when it does not find sbatch
        with open(os.devnull, "wb") as devnull:
            sp.check_output(command, stderr=devnull).decode()
        # if it runs without error, we have access to slurm
        return True
    # if an error is raised we do not have slurm
    except (sp.CalledProcessError, FileNotFoundError):
        return False


@cache.timed_cache(3)
def squeue() -> results.Result:
    """
    Get information about jobs managed by slurm using squeue.

    Returns:
        : A :class:`Result <tcutility.results.result.Result>` object containing information about the calculation status:

            - ``directory`` **(list[str])** – path to slurm directories.
            - ``id`` **(list[str])** – slurm job id's.
            - ``status`` **(list[str])** – slurm job status name. See squeue documentation.
            - ``statuscode`` **(list[str])** – slurm job status codes. See squeue documentation

    .. note::

        By default this function uses a timed cache (see :func:`timed_cache <tcutility.cache.timed_cache>`) with a 3 second delay to lessen the load on HPC systems.
    """
    ret = results.Result()

    if not has_slurm():
        return ret

    # specify the columns to get here
    columns = ["directory", "id", "statuscode", "status"]
    options = ["%Z", "%A", "%t", "%T"]  # these are the squeue format codes

    # set each column as an empty list in the return object
    for col in columns:
        ret[col] = []

    # run the squeue command with the formatting options
    output = sp.check_output(["squeue", "--me", "--format", "" + " ".join(options) + ""]).decode()
    output = [line for line in output.splitlines()[1:] if line.strip()]

    # then add the data to the return object's lists
    for line in output:
        [ret[col].append(val) for col, val in zip(columns, line.split())]

    return ret


def sbatch(runfile: str, **options: dict) -> results.Result:
    """
    Submit a job to slurm using sbatch.

    Args:
        runfile: the path to the filename to be submitted.
        options: options to be used for sbatch.

    Returns:
        : A :class:`Result <tcutility.results.result.Result>` object containing information about the newly submitted slurm job

            - ``id`` **(str)** - the ID for the submitted slurm job.
            - ``command`` **(str)** - the command used to submit the job.
    """
    cmd = "sbatch "
    for key, val in options.items():
        key = key.replace("_", "-")

        if val is True:
            if len(key) > 1:
                cmd += f"--{key} "
            else:
                cmd += f"-{key} "
        else:
            if len(key) > 1:
                cmd += f"--{key}={val} "
            else:
                cmd += f"-{key} {val} "

    cmd = cmd + runfile

    ret = results.Result()
    ret.command = cmd

    # run the job
    sbatch_out = sp.check_output(cmd.split(), stderr=sp.STDOUT).decode()
    # get the slurm job id from the output
    for line in sbatch_out.splitlines():
        if "Submitted batch job" in line:
            # set the slurm job id for this calculation, we use this in order to set dependencies between jobs.
            ret.id = line.strip().split()[-1]
            break

    return ret


def workdir_info(workdir: str) -> results.Result:
    """
    Function that gets squeue information given a working directory. This will return None if the directory is not being actively referenced by slurm.

    Returns:
        :Result object containing information about the calculation status, see :func:`squeue`.
    """
    if not has_slurm():
        return None

    sq = squeue()
    if workdir not in sq.directory:
        return None

    workdir_index = sq.directory.index(workdir)
    ret = results.Result()
    for key, vals in sq.items():
        ret[key] = vals[workdir_index]

    return ret


def wait_for_job(slurmid: int, check_every: int = 3):
    """
    Wait for a slurm job to finish. We check every `check_every` seconds if the slurm job id is still present in squeue.

    Args:
        slurmid: the ID of the slurm job we are waiting for.
        check_every: the amount of seconds to wait before checking squeue again. Don't put this too high, or you will anger the cluster people.
    """
    while slurmid in squeue().id:
        time.sleep(check_every)


if __name__ == "__main__":
    if has_slurm():
        log.info("This platform has SLURM.")
    else:
        log.info("This platform does not have SLURM.")



import os
from enum import Enum, auto


class OSName(Enum):
    """
    An enumeration of the different operating systems.
    """

    WINDOWS = auto()
    LINUX = auto()
    MACOS = auto()


def get_os_name() -> OSName:
    """
    Get the name of the operating system. Returns a value from the :class:`OSName <tcutility.environment.OSName>` enumeration.
    """
    os_name = os.name
    if os_name == "nt":
        return OSName.WINDOWS
    elif os_name == "posix":
        return OSName.LINUX
    elif os_name == "mac":
        return OSName.MACOS
    else:
        raise ValueError(f"Unknown operating system: {os_name}")



if __name__ == '__main__':
	from tcutility import results, molecule, constants
	import os

	# load this calculation
	res = results.read(os.getcwd())

	# check if the calculation succeeded
	if res.status.fatal:
		exit()

	os.makedirs('converged_mols', exist_ok=True)
	# get the molecules that converged
	converged_mols = [mol for mol, converged in zip(res.history.molecule, res.history.converged) if converged]
	energies = [energy for energy, converged in zip(res.history.energy, res.history.converged) if converged]
	# write the molecules to a file
	with open('ams.amv', 'w+') as amv:
		# the molecules should not have a header like a regular amv file, but only the coordinates
		for i, (mol, energy) in enumerate(zip(converged_mols, energies)):
			for atom in mol:
				amv.write(f'{atom.symbol} {atom.x} {atom.y} {atom.z}\n')
			
			# molecules are separated by empty lines
			amv.write('\n')

			molecule.save(mol, f'converged_mols/{i}.xyz', comment=f'E = {energy * constants.HA2KCALMOL:.6f} kcal/mol')

			if energy == max(energies):
				molecule.save(mol, 'converged_mols/highest_energy.xyz', comment=f'E = {energy * constants.HA2KCALMOL:.6f} kcal/mol')

			if energy == min(energies):
				molecule.save(mol, 'converged_mols/lowest_energy.xyz', comment=f'E = {energy * constants.HA2KCALMOL:.6f} kcal/mol')



import pathlib as pl
from typing import List

import click
from tcutility.analysis.task_specific.irc import concatenate_irc_trajectories, create_result_objects
from tcutility.log import log
from tcutility.molecule import write_mol_to_amv_file, write_mol_to_xyz_file


@click.command()
@click.argument("jobs", nargs=-1, type=click.Path(exists=True))
@click.option("-r", "--reverse", is_flag=True, help="Reverses the trajectory")
@click.option("-o", "--output", type=click.Path(), default="./", help="Directory in which the output file will be saved")
@click.option("-l", "--log_level", type=int, default=20, help="Set the log level. The lower the value, the more is printed. Default is 20 (info).")
def concatenate_irc_paths(jobs: List[str], reverse: bool, output: str, log_level: int) -> None:
    """
    Combine separated IRC paths.

    Scripts that takes in two or more directories containing an IRC file ("ams.rkf") and concatenates them through the RMSD values. Produces a .xyz and .amv file in the specified output directory.
    The output directory is specified with the -o flag. If not specified, the output will be written to the current working directory.
    In addition, the -r flag can be used to reverse the trajectory.

    Note: ALWAYS visualize the .amv file in AMSView to verify the trajectory.
    """
    outputdir = pl.Path(output).resolve()
    job_dirs = [pl.Path(directory).resolve() for directory in jobs]

    if len(job_dirs) > 2:
        raise ValueError("Only two IRC paths can be concatenated at a time as is currently implemented.")

    log(f"Concatenating trajectories... with reverse = {reverse}", log_level)
    res_objects = create_result_objects(job_dirs)
    molecules, energies = concatenate_irc_trajectories(res_objects, reverse)
    log(f"Trajectories concatenated successfully (total length = {len(molecules)}) .", log_level)

    outputdir.mkdir(parents=True, exist_ok=True)
    write_mol_to_amv_file(out_file=outputdir / "concatenated_mols.amv", mols=molecules, energies=energies)
    write_mol_to_xyz_file(out_file=outputdir / "concatenated_mols.xyz", mols=molecules)

    log(f"Output written to {outputdir / 'concatenated_mols'}", log_level)


if __name__ == "__main__":
    concatenate_irc_paths()



from tcutility import results
from tcutility.analysis.vibration import ts_vibration
import pytest
import numpy as np
from pathlib import Path


@pytest.fixture
def sn2_path() -> Path:
    return Path(__file__).parent / "fixtures" / "chloromethane_sn2_ts"


@pytest.fixture
def sn2_result() -> results.Result:
    return results.read(Path(__file__).parent / "fixtures" / "chloromethane_sn2_ts")


@pytest.fixture
def rad_path() -> Path:
    return Path(__file__).parent / "fixtures" / "radical_addition_ts"


@pytest.fixture
def rad_result() -> results.Result:
    return results.read(Path(__file__).parent / "fixtures" / "radical_addition_ts")


def test_sn2_determine_rc(sn2_result) -> None:
    assert np.array_equal(ts_vibration.determine_ts_reactioncoordinate(sn2_result), np.array([[1, 2, 1], [1, 6, -1]]))


def test_rad_determine_rc_1(rad_result) -> None:
    assert np.array_equal(ts_vibration.determine_ts_reactioncoordinate(rad_result, min_delta_dist=0.1), np.array([[1, 16, -1]]))


def test_rad_determine_rc_2(rad_result) -> None:
    assert np.array_equal(ts_vibration.determine_ts_reactioncoordinate(rad_result, min_delta_dist=0.0), np.array([[1, 2, 1], [1, 8, 1], [1, 16, -1], [8, 9, -1]]))


def test_validate_sn2_ts_1(sn2_path) -> None:
    result = ts_vibration.validate_transitionstate(sn2_path, [[1, 2, 1], [1, 6, -1]])
    assert result is True


def test_validate_sn2_ts_2(sn2_path) -> None:
    result = ts_vibration.validate_transitionstate(sn2_path, [[2, 3, 1], [1, 6, -1]])
    assert result is False  # [2,3,1] should not be present


def test_validate_rad_ts_1(rad_path) -> None:
    result = ts_vibration.validate_transitionstate(rad_path)
    assert result is True


def test_validate_rad_ts_2(rad_path) -> None:
    result = ts_vibration.validate_transitionstate(rad_path, [1, 16, 1], min_delta_dist=0.1)
    assert result is True


def test_validate_rad_ts_3(rad_path) -> None:
    result = ts_vibration.validate_transitionstate(rad_path, [[9, 8, -3], [16, 1, -5]], min_delta_dist=0.0)
    assert result is True  # Testing user input normalization


def test_validate_rad_ts_4(rad_path) -> None:
    result = ts_vibration.validate_transitionstate(rad_path, [1, 16])
    assert result is True


def test_validate_rad_ts_5(rad_path) -> None:
    result = ts_vibration.validate_transitionstate(rad_path, [8, 9, 1], min_delta_dist=0.1)
    assert result is False  # Should not be present with this min_delta_dist parameter value


def test_validate_rad_ts_6(rad_path) -> None:
    result = ts_vibration.validate_transitionstate(rad_path, [[1, 16, 1], [8, 9, -1]])
    assert result is False  # Should have equal signs


def test_validate_rad_ts_7(rad_path) -> None:
    # Sign has to be provided for all reaction coordinates or none of them. ValueError should be raised otherwise
    with pytest.raises(ValueError):
        ts_vibration.validate_transitionstate(rad_path, [[1, 16], [8, 9, 1]])



""" Module containing CLI functionality for resizing pictures containing molecules """
from tcutility import report
import os
import tkinter as tk
from tkinter import filedialog
import click
from pathlib import Path


root = tk.Tk()
root.withdraw()


@click.command()
@click.option('-f', '--folder', type=click.Path(exists=True))
@click.option('-p', '--padding', default='10%', help="""The amount of padding to add to the resized figures.
If given an integer we use pixel padding. E.g. -p 50 will add a padding of 50 pixels.
Add a %%-sign to use relative padding. E.g. -p 10%% will add a padding of 10%%.""")
def resize(folder: Path, padding: str):
    """Resize images containing molecules.

    This CLI-program resizes images in a directory based on detected circles. It will ensure the selected circles are placed at the same location and are also resized to be the same size.
    When starting the program it will show you for each image numbered detected atoms. 
    Entering the desired number into the CLI will select it for resizing. If you do not enter a number, the figure will be ignored for further processing.
    New images will be written to the folder postpended with _fixed.
    """
    circle_numbers = {}
    if folder is not None:
        img_paths = [os.path.join(folder, img_path) for img_path in os.listdir(folder)]
    else:
        img_paths = filedialog.askopenfilenames()

    for img_path in img_paths:
        report.figure_resizer._analyse_img(img_path, plot=True)
        circle = input(f'Select circle for {img_path}, leave empty to skip: ')
        if circle == '':
            continue

        circle_numbers[img_path] = int(circle)

    report.figure_resizer.resize(img_paths, circle_numbers, padding=padding)



from tcutility import results


def test_init():
    res = results.result.Result()  # noqa F841


def test_assign():
    res = results.result.Result()
    res.a = 10
    assert res.a == 10


def test_assign2():
    res = results.result.Result()
    res.a.b = 10
    assert res.a.b == 10


def test_assign3():
    res = results.result.Result()
    res.a.b = 10
    res.a.c = 20
    assert res.a.b == 10 and res.a.c == 20


def test_assign4():
    res = results.result.Result()
    res.a.lst = []
    res.a.lst.append(10)
    assert len(res.a.lst) == 1


def test_contains():
    res = results.result.Result()
    res.a = 10
    assert "a" in res


def test_contains2():
    res = results.result.Result()
    res.a.b = 10
    assert "a" in res


def test_contains3():
    res = results.result.Result()
    res.a.b = 10
    assert "b" in res.a


def test_contains_neg():
    res = results.result.Result()
    assert "a" not in res


def test_contains_neg2():
    res = results.result.Result()
    assert "b" not in res.a


def test_contains_neg3():
    res = results.result.Result()
    res.a.b = 10
    assert "b" not in res


def test_contains_neg4():
    res = results.result.Result()
    res.a
    assert "a" not in res



import os
from tcutility import results

j = os.path.join


def test_reading() -> None:
    input_dir = j(os.path.split(__file__)[0], "fixtures", "input_files")
    for file in os.listdir(input_dir):
        with open(j(input_dir, file)) as inp:
            results.ams.get_ams_input(inp.read())



import pathlib as pl

import pytest
from tcutility.analysis.task_specific.irc import _get_converged_molecules, concatenate_irc_trajectories
from tcutility.results import read
from tcutility.results.result import Result

#
current_dir = pl.Path(__file__).parent
ircs_dir = current_dir / "fixtures" / "irc_trajectories"


@pytest.fixture
def forward_irc() -> Result:
    return read(ircs_dir / "forward")


@pytest.fixture
def forward2_irc() -> Result:
    return read(ircs_dir / "forward_2")


@pytest.fixture
def backward_irc() -> Result:
    return read(ircs_dir / "backward")


@pytest.fixture
def backward2_irc() -> Result:
    return read(ircs_dir / "backward_2")


def test_len_irc_trajectories(forward_irc, forward2_irc, backward_irc, backward2_irc):
    res_objs = [forward_irc, forward2_irc, backward_irc, backward2_irc]

    expected_lengths = [84, 88, 91, 85]

    for expected_length, res_obj in zip(expected_lengths, res_objs):
        assert len(_get_converged_molecules(res_obj)) == expected_length


def test_len_final_trajectory(forward_irc, backward_irc, backward2_irc, forward2_irc):
    expected_lengths = [84, 91, 85, 88]

    concatenated_mols, _ = concatenate_irc_trajectories([forward_irc, backward_irc, backward2_irc, forward2_irc])
    assert (
        len(concatenated_mols) == sum(expected_lengths) - 2
    )  # 2 is the number of duplicated points in the concatenated trajectory. In this case, two seperate ircs are concatenated and so there are two duplicated points.

    concatenated_mols, _ = concatenate_irc_trajectories([forward_irc, backward_irc], reverse=True)
    assert len(concatenated_mols) == 84 + 91 - 1


if __name__ == "__main__":
    pytest.main()



from tcutility.results import cache, Result
from tcutility import ensure_list, constants


def get_calc_settings(info: Result) -> Result:
    """
    Function to read useful calculation settings from kf reader
    """

    assert info.engine == "dftb", f"This function reads DFTB data, not {info.engine} data"

    if info.input.task == 'PESScan':
        return Result()

    assert "dftb.rkf" in info.files, f'Missing dftb.rkf file, [{", ".join([": ".join(item) for item in info.files.items()])}]'

    ret = Result()

    # set the calculation task at a higher level
    ret.task = info.input.task

    return ret


def get_properties(info: Result) -> Result:
    """Function to get properties from an DFTB calculation.

    Args:
        info: Result object containing DFTB properties.

    Returns:
        :Result object containing properties from the DFTB calculation:

            - **energy.bond (float)** – bonding energy (|kcal/mol|).
    """

    assert info.engine == "dftb", f"This function reads DFTB data, not {info.engine} data"

    if info.input.task == 'PESScan':
        return Result()

    assert "dftb.rkf" in info.files, f'Missing dftb.rkf file, [{", ".join([": ".join(item) for item in info.files.items()])}]'

    reader_dftb = cache.get(info.files["dftb.rkf"])
    ret = Result()

    # properties are not given as in ADF. Instead you must first know which index each key is at. For example we might have:
    # (Properties, SubType(k)) == Coulomb
    # therefore, to get the Coulomb energy we first have to know the index k
    # we therefore first loop through every subtype entry and store the substypes in order in a list
    # we also store the main type (e.g. 'Energy', 'Gradient', ...) and the value
    number_of_properties = reader_dftb.read("Properties", "nEntries")
    subtypes = []
    types = []
    values = []
    for i in range(1, number_of_properties + 1):
        subtypes.append(reader_dftb.read("Properties", f"Subtype({i})").strip())
        types.append(reader_dftb.read("Properties", f"Type({i})").strip())
        if types[-1] == 'Energy':
            values.append(reader_dftb.read("Properties", f"Value({i})") * constants.HA2KCALMOL)
        else:
            values.append(reader_dftb.read("Properties", f"Value({i})"))

    # then simply add the properties to ret
    for typ, subtyp, value in zip(types, subtypes, values):
        ret[typ.replace(" ", "_")][subtyp] = value

    if ret.energy['DFTB Final Energy']:
        ret.energy.bond = ret.energy['DFTB Final Energy']

    # we also read vibrations
    if ('Vibrations', 'nNormalModes') in reader_dftb:
        ret.vibrations.number_of_modes = reader_dftb.read("Vibrations", "nNormalModes")
        ret.vibrations.frequencies = ensure_list(reader_dftb.read("Vibrations", "Frequencies[cm-1]"))
        if ("Vibrations", "Intensities[km/mol]") in reader_dftb:
            ret.vibrations.intensities = ensure_list(reader_dftb.read("Vibrations", "Intensities[km/mol]"))
        ret.vibrations.number_of_imag_modes = len([freq for freq in ret.vibrations.frequencies if freq < 0])
        ret.vibrations.character = "minimum" if ret.vibrations.number_of_imag_modes == 0 else "transitionstate"
        ret.vibrations.modes = []
        for i in range(ret.vibrations.number_of_modes):
            ret.vibrations.modes.append(reader_dftb.read("Vibrations", f"NoWeightNormalMode({i+1})"))

    if ("Thermodynamics", "Gibbs free Energy") in reader_dftb:
        ret.energy.gibbs = reader_dftb.read("Thermodynamics", "Gibbs free Energy") * constants.HA2KCALMOL
        ret.energy.enthalpy = reader_dftb.read("Thermodynamics", "Enthalpy") * constants.HA2KCALMOL
        ret.energy.nuclear_internal = reader_dftb.read("Thermodynamics", "Internal Energy total") * constants.HA2KCALMOL

    return ret



"""
Tiny module that handles caching rkf files.
rkf files take a long time to open (especially {engine}.rkf), so it is better to open them once and cache them for later use
"""
from typing import Any, Union

from scm import plams

# the actual cache is stored in this dict
_cache = {}


class TrackKFReader(plams.KFReader):
    """Subclass of plams.KFReader that also tracks the variables that were read. This class can be useful to figure out which variables are important.
    For example, we can then trim rkf files to reduce their filesizes."""

    def __init__(self, *args, **kwargs):
        self.tracker = []
        super().__init__(*args, **kwargs)

    def read(self, section: str, variable: str) -> Any:
        """Read a variable from a section of an rkf file and store the accessed section and variable in the tracker.

        Args:
            section: Name of the section to read from.
            variable: Name of the variable inside the section.

        Returns:
            Any value returned will be automatically converted into the correct type. E.g. text-based data will be converted to a string, numbers to floats or integers, etc."""

        self.tracker.append((section, variable))
        return super().read(section, variable)


def store(reader: plams.KFReader) -> None:
    """Store an rkf reader in the cache. It can later be indexed by its path.

    Args:
        reader: An rkf file reader object."""
    path = reader.path
    _cache[path] = reader


def get(path: str) -> plams.KFReader:
    """Retrieve an rkf reader from storage using its path. If the file was not opened yet, open it first and then store and return the new object.

    Args:
        path: path to the rkf file location.

    Returns:
        An rkf file reader that can be used for reading data from a calculation.
    """
    if not path:
        return None

    # if the path is already in the cache, simply return it
    if path in _cache:
        return _cache[path]
    # else we will load the rkf file and store it in the cache
    reader = TrackKFReader(path)
    store(reader)
    return reader


def unload(arg: Union[str, plams.KFReader]):
    """Delete an rkf reader from storage. Since rkf files can be quite large, we should not forget to unload them after we have finished reading from them, lest we run into memory issues.

    Args:
        arg: Either a path pointing to the rkf file location or the reader itself."""
    # if the arg is a key in the cache we will delete the key
    if arg in _cache:
        del _cache[arg]
        return
    # if it is not a key, find the key corresponding to the value and delete it
    if arg in _cache.values():
        key = [key for key, val in _cache.items() if val == arg][0]
        del _cache[key]
        return



import inspect
import json
import os
import sys
from datetime import datetime
from math import ceil
from time import perf_counter
from typing import Any, Generator, Iterable, List, Sequence, TypeVar, Union

import numpy as np

from tcutility import ensure_2d

# from threading import Thread


###########################################################
# MODULE LEVEL VARIABLES USED TO CHANGE LOGGING BEHAVIOUR #
logfile = sys.stdout  # stream or file to send prints to. Default sys.stdout is the standard Python print output
errfile = sys.stderr  # stream or file to send prints to. Default sys.stderr is the standard Python print output
tab_level = 0  # number of tabs to add before a logged message. This is useful to build a sense of structure in a long output
max_width = 0  # maximum width of printed messages. Default 0 means unbounded
print_date = True  # print time stamp before a logged message
log_level = 20  # the level of verbosity. Messages with a log_level >= this number will be sent to logfile


class Emojis:
    """
    Class containing some useful emojis and other characters.
    Supports dot-notation and indexation to get a character.

    E.g. ``Emojis.wait == Emojis['wait']``
    """

    wait = "🕒"
    good = "✅"
    cancel = "🛑"
    sleep = "💤"
    fail = "❌"
    send = "📤"
    receive = "📥"
    empty = "⠀⠀"
    finish = "🏁"
    warning = "⚠️"
    question = "❔"
    info = "ℹ️"
    rarrow = "─>"
    larrow = "<─"
    lrarrow = rlarrow = "<─>"
    angstrom = "Å"

    def __init__(self):
        raise SyntaxError("Do not instantiate the Emojis-class.")

    def __class_getitem__(cls, item):
        return getattr(cls, item)


class NoPrint:
    """
    Context-manager that suppresses printing. It works by redirecting prints to a temporary output file.
    This file is deleted after exiting the context-manager.
    """

    def __init__(self, stdout=None, stderr=None):
        self.stdout = stdout or logfile
        self.stderr = stderr or errfile
        self.overflow = open(str(os.getpid()) + "_NOPRINT.tmp", "w+")

    def __enter__(self):
        global logfile, errfile
        logfile = self.overflow
        errfile = self.overflow

    def __exit__(self, *args, **kwargs):
        global logfile, errfile
        logfile = self.stdout
        errfile = self.stderr
        self.overflow.close()
        os.remove(str(os.getpid()) + "_NOPRINT.tmp")


def time_stamp():
    """
    Return the current timestamp in a "[YYYY/MM/DD HH:MM:SS] "" format.
    """
    now = datetime.now()
    return f"[{now.year}/{str(now.month).zfill(2)}/{str(now.day).zfill(2)} {str(now.hour).zfill(2)}:{str(now.minute).zfill(2)}:{str(now.second).zfill(2)}] "


def log(message: Any = "", level: int = 20, end: str = "\n"):
    r"""
    Print a nicely formatteed message.
    This function adds the current timestamp and supports multi-line printing (split on the ``\n`` escape character).
    For verbosity levels we use the following convention:

    .. code-block::

        NOTSET   = 0
        DEBUG    = 10
        INFO     = 20
        WARN     = 30
        ERROR    = 40
        CRITICAL = 50


    Args:
        message: the message to send. Before printing we will use the ``message.__str__`` method to get the string representation. If the message is a ``dict`` we use the ``json`` module to format the message nicely.
        level: the level to print the message at. We compare the level against the module-wide ``log_level`` variable (by default ``log_level = 20``). If the level is below ``log_level`` we do not print it.
        end: the end of the string. This is usually the new-line character ``\n``.
    """
    if level < log_level:
        return

    # dictionaries are handled specially so that they look nice
    if isinstance(message, dict):
        message = json.dumps(message, indent=4, sort_keys=True)

    # Print each line separately
    splitted_message = str(message).split("\n")
    for m in splitted_message:
        # handle lines that exceed the maximum print width
        if max_width > 0 and len(m) > max_width:
            m = m[: max_width - 4] + " ..."

        # add the tab-levels and timestamp
        m = "\t" * tab_level + m
        if print_date:
            m = time_stamp() + m

        print(m, file=logfile, end=end, flush=True)


def flow(message: str = "", tags: List[str] = ["straight"], level: int = 20) -> None:
    """
    Function to create flowchart-like output.
    It will print a message prepended by flow elements (arrows and lines).
    The flow elements are determined based on the given tags.
    """
    elements = {
        "start": "┯ ",
        "startinv": "┷ ",
        "end": "╰─> ",
        "straight": "│   ",
        "split": "├─> ",
        "skip": "    ",
        "vert": "────",
        "endvert": "╰──> ",
        "splitvert": "├──> ",
    }
    s = ""
    for tag in tags:
        s += elements[tag]
    log(s + message, level=level)


def table(rows: List[List[Any]], header: Union[List[str], None] = None, sep: str = "   ", hline: List[int] = [], level: int = 20) -> str:
    r"""
    Print a table given rows and a header. Values in rows will be cast to strings first.

    Args:
        rows: list of `nrows` sequences containing `ncols` data inside the table.
        header: list of `ncols` strings that represent the column names of the table. They will be printed at the top of the table.
        sep: str representing the separation between columns.
        hline: list of integers specifying rows after which lines will be drawn. Supports negative indices, e.g. -1 will draw a line at the end of the table.

    Returns:
        str: the table in string format, where lines are separated by "\n"
    """
    rows = ensure_2d(rows)
    nrows = len(rows)
    ncols = len(rows[0])

    hline = [h + nrows + 1 if h < 0 else h for h in hline]

    if header:
        rows = [header] + rows
        hline = [0] + hline

    rows = [[str(x) for x in row] for row in rows]

    column_sizes = [max(len(row[i]) for row in rows) for i in range(ncols)]
    return_str = ""
    for i, row in enumerate(rows):
        row_str = sep.join([x.ljust(column_size) for x, column_size in zip(row, column_sizes)]) + "\n"
        return_str += row_str
        if i in hline:
            return_str += "─" * len(row_str) + "\n"

    log(return_str, level=level)
    return return_str


def rectangle_list(values: Sequence, spaces_before: int = 0, level: int = 20):
    """
    This function prints a list of strings in a rectangle to the output.
    This is similar to what the ls program does in unix.
    """
    n_shell_col = os.get_terminal_size().columns
    # we first have to determine the correct dimensions of our rectangle
    for ncol in range(1, n_shell_col):
        # the number of rows for the number of columns
        nrows = ceil(len(values) / ncol)
        # we then get what the rectangle would be
        mat = [[str(x) for x in values[i * ncol : (i + 1) * ncol]] for i in range(nrows)]
        # and determine for each column the width
        col_lens = [max([len(row[i]) for row in mat if i < len(row)] + [0]) for i in range(ncol)]
        # then calculate the length of each row based on the column lengths
        # we use a spacing of 2 spaces between each column
        row_len = 22 * print_date + spaces_before + sum(col_lens) + 2 * len(col_lens) - 2

        # if the rows are too big we exit the loop
        if row_len > n_shell_col:
            break

        # store the previous loops results
        prev_col_lens = col_lens
        prev_mat = mat

    # then print the strings with the right column widths
    for row in prev_mat:
        log(" " * spaces_before + "  ".join([x.ljust(col_len) for x, col_len in zip(row, prev_col_lens)]), level=level)


T = TypeVar("T")


def loadbar(sequence: Union[Iterable[T], Sequence[T]], comment: str = "", Nsegments: int = 50, Nsteps: int = 10, level: int = 20) -> Generator[T, None, None]:
    """
    Return values from an iterable ``sequence`` and also print a progress bar for the iteration over this sequence.

    Args:
        sequence: any iterable sequence. Should define the ``__len__`` method.
        comment: a string to be printed at the end of the loading bar to give information about the loading bar.
        Nsegments: length of the loading bar in characters.
        Nsteps: number of times to print the loading bar during iteration. If the output is a tty-type stream Nsteps will be set to the length of sequence.
    """
    if not isinstance(sequence, Sequence):
        chars = ["⠀", "⠄", "⠆", "⠦", "⠧", "⠷", "⠿", "⠻", "⠛", "⠙", "⠉", "⠈", "⠀"]
        iteration = 0
        if not logfile.isatty() and comment:
            log(comment, level=level)

        starttime = perf_counter()
        for val in sequence:
            iteration += 1
            elapsed_time = perf_counter() - starttime
            time_per_step = elapsed_time / iteration
            # every 0.1 seconds we change the character
            char_step = int(elapsed_time / 0.1) % len(chars)
            if logfile.isatty():
                log(f"{chars[char_step]} {comment} [Steps: {iteration}, Elapsed: {elapsed_time:.1f}s]", end="\r", level=level)

            yield val

        log(level=level)
        return

    N = len(sequence)
    # if the output stream is tty-type we set the number of steps to the lenth of the sequence so the loading bar looks smoother
    Nsteps = N if logfile.isatty() else Nsteps
    Ndigits = int(np.log10(N)) + 1  # well-known method to get number of digits of an integer
    # we track what the maximum length of the loading bar is.
    # We use the '\r' return carriage when logging, so we have to overwrite the whole previous line.
    # For this we must know the largest length of our loading bar strings.
    max_length = 0

    # track when the loading bar started. We use this to calculate the ETA later
    loading_bar_start_time = perf_counter()
    for i, val in enumerate(sequence):
        # we only print the loading bar on every Nsteps iterations
        if i % (N // min(N, Nsteps)) != 0:
            yield val
            continue

        # calculate how many segments should be filled and empty
        Nfilled_segments = int(i / N * Nsegments)
        filled = "─" * Nfilled_segments
        empty = " " * (Nsegments - Nfilled_segments)

        # determine what the cursor should look like
        if Nfilled_segments == Nsegments:  # if the bar is fully filled it should be just a stripe
            cursor = "─"
        elif i == 0:  # if the bar is empty we don't draw the cursor
            cursor = " "
        else:  # normal case
            cursor = ">"

        # calculate the ETA. In the first iteration we cannot know yet
        if i == 0:
            eta = "???"
        else:
            # the algorithm simply extrapolates the current time taken to the case of the final iteration
            time_taken = perf_counter() - loading_bar_start_time
            time_per_step = time_taken / i
            time_left = time_per_step * (N - i)
            eta = f"{time_left:.1f}"

        # build the loading bar string
        s = f'{i:{Ndigits}}/{N} {"├" if i > 0 else "|"}{filled + cursor + empty}{"│"} {i/N:4.0%} {comment} [ETA: {eta}s]'.ljust(max_length)

        # update the max loading bar string length
        max_length = max(len(s), max_length)
        log(s, end="\r", level=level)

        yield val

    # plot the final iteration at 100% filled
    s = f'{i+1:{Ndigits}}/{N} {"├"}{"─"*(Nsegments+1)}┤ 100% {comment}'.ljust(max_length)
    log(s, end="\r", level=level)
    log(level=level)


def boxed(message: str, title: Union[str, None] = None, message_align: str = "left", title_align: str = "left", round_corners: bool = True, double_edge: bool = False, level: int = 20) -> None:
    r"""
    Print a message surrounded by a box with optional title.

    Args:
        message: the message to place in the box. Multiline messages are separated by "\n".
        title: the title placed in the top edge of the box.
        message_align: alignment of the text inside the box. One of ["left", "center", "right"].
        title_align: alignment of the title. One of ["left", "center", "right"].
        round_corners: whether the corners of the box should be rounded or not. Rounded corners are only available for single-edge boxes.
        double_edge: whether the edges of the box should be double.

    Returns:
        The printed message in strings format.
    """
    # define the edges and corners
    straights = ["│", "─"]
    if round_corners and not double_edge:
        corners = ["╭", "╮", "╯", "╰"]
    elif double_edge:
        corners = ["╔", "╗", "╝", "╚"]
        straights = ["║", "═"]
    else:
        corners = ["┌", "┐", "┘", "└"]

    # get the width the box should have
    messages = message.split("\n")
    # the length is influenced by both the messages as well as the title
    maxlen = max(max(len(message) for message in messages), len(title or ""))

    # build first row containing the title
    if title is not None:
        if title_align == "left":
            s = corners[0] + (" " + title + " ").ljust(maxlen + 2, straights[1]) + corners[1] + "\n"
        elif title_align == "right":
            s = corners[0] + (" " + title + " ").rjust(maxlen + 2, straights[1]) + corners[1] + "\n"
        else:
            s = corners[0] + (" " + title + " ").center(maxlen + 2, straights[1]) + corners[1] + "\n"
    else:
        s = corners[0] + straights[1] * (maxlen + 2) + corners[1] + "\n"

    # build main body of box
    for message in messages:
        if message_align == "left":
            s += f"{straights[0]} " + message.ljust(maxlen) + f" {straights[0]}\n"
        if message_align == "right":
            s += f"{straights[0]} " + message.rjust(maxlen) + f" {straights[0]}\n"
        if message_align == "center":
            s += f"{straights[0]} " + message.center(maxlen) + f" {straights[0]}\n"

    # build bottom row
    s += corners[3] + straights[1] * (maxlen + 2) + corners[2]

    log(s, level=level)


def debug(message: str, level: int = 10, caller_level: int = 2):
    """
    Print a debug message.
    """
    log(f"[DEBUG]({caller_name(caller_level)}): " + message, level=level)


def info(message: str, level: int = 20, caller_level: int = 2):
    """
    Print an informative message.
    """
    log(f"[INFO]({caller_name(caller_level)}): " + message, level=level)


def warn(message: str, level: int = 30, caller_level: int = 2):
    """
    Print a warning message.
    """
    log(f"[WARNING]({caller_name(caller_level)}): " + message, level=level)


def error(message: str, level: int = 40, caller_level: int = 2):
    """
    Print an error message.
    """
    log(f"[ERROR]({caller_name(caller_level)}): " + message, level=level)


def critical(message: str, level: int = 50, caller_level: int = 2):
    """
    Print a critical message.
    """
    log(f"[CRITICAL]({caller_name(caller_level)}): " + message, level=level)


def caller_name(level: int = 1) -> str:
    """
    Return the full name of the caller of a function.

    Args:
        level: the number of levels to skip when getting the caller name. Level 1 is always this function. When used by a different function it should be set to 2. E.g. when using the log.warn function level is set to 2.

    Returns:
        The full name of the caller function.
    """
    stack = inspect.stack()
    names = [stack[level][3]]  # add the original function, this should be at a certain levels, since this function is called by warn or info or error
    for frame in stack[level:]:
        if frame.function == "<module>":  # <module> is always there, so we can skip it
            continue

        # get the more interesting names
        # first check if the
        if hasattr(frame[0], "f_locals") and "self" in frame[0].f_locals:
            clas = frame[0].f_locals["self"].__class__
            module = clas.__module__
            if module == "builtins":
                names.append(clas.__qualname__)
            else:
                names.append(module + "." + clas.__qualname__)
        elif hasattr(frame[0], "f_code"):
            names.append(frame[0].f_code.co_name)
    names_ordered = []
    for name in names[::-1]:
        if name not in names_ordered:
            names_ordered.append(name)
    return ".".join(names_ordered)


if __name__ == "__main__":
    boxed("testing, 1, 2, 3, 4, 5, 6, 7, 8\nsecond row, 1, 2, 3, 4, 5, 6\n...\n...\n\n...\n...\nLast row here", title="ReactionRunner")

    with NoPrint():
        log("test test test")
        log("test test test")
        log("test test test")
        log("test test test")

    log()

    flow("Start", ["start"])
    flow()
    flow("First step", ["split"])
    flow("First substep of first step", ["straight", "split"])
    flow("Second substep of first step", ["straight", "split"])
    flow("Third and final substep of first step", ["straight", "end"])
    flow()
    flow("Second step", ["split"])
    flow("Substep of second step", ["straight", "split"])
    flow("Substep of substep of second step", ["straight", "straight", "end"])
    flow("Substep of substep of substep of second step", ["straight", "straight", "skip", "end"])
    flow("Substep of second step", ["straight", "end"])
    flow()
    flow("Final step", ["split"])
    flow()
    flow(f"{Emojis.good} The end", ["startinv"])

    log()

    info("This is important info")
    warn("This is an important warning!")

    import time

    for x in loadbar(range(100), "Sleeping test"):
        time.sleep(0.01)

    x = np.linspace(1, 12, 12)
    rows = np.vstack([x, x**2, x**3, x**4, x**5]).astype(int).T.tolist()
    table(rows, header=["X", "y=x^2", "y=x^3", "y=x^4", "y=x^5"], hline=[-1])

    from tcutility import log as cli_log

    class TestClass:
        def test_method(self):
            cli_log.warn("I am testing the warning function")

    TestClass().test_method()

    rectangle_list(range(1000))



import os
import re
from typing import Dict, List

from tcutility import results

j = os.path.join


def split_all(path: str) -> List[str]:
    """
    Split a path into all of its parts.

    Args:
        path: the path to be split, it will be separated using :func:`os.path.split`.

    Returns:
        A list of parts of the original path.

    Example:
        .. code-block:: python

            >>> split_all('a/b/c/d')
            ['a', 'b', 'c', 'd']
    """
    path = os.path.normpath(path)
    parts = []
    while True:
        a, b = os.path.split(path)
        if not a or not b:
            parts.append(path)
            return parts[::-1]
        parts.append(b)
        path = a


def get_subdirectories(root: str, include_intermediates: bool = False) -> List[str]:
    """
    Get all sub-directories of a root directory.

    Args:
        root: the root directory.
        include_intermediates: whether to include intermediate sub-directories instead of only the lowest levels.

    Returns:
        A list of sub-directories with ``root`` included in the paths.

    Example:
        Given a file-structure as follows:

        .. code-block::

            root
            |- subdir_a
            |  |- subsubdir_b
            |  |- subsubdir_c
            |- subdir_b
            |- subdir_c

        Then we get the following outputs.

        .. tabs::

            .. group-tab:: Including intermediates

                .. code-block:: python

                    >>> get_subdirectories('root', include_intermediates=True)
                    ['root',
                     'root/subdir_a',
                     'root/subdir_a/subsubdir_b',
                     'root/subdir_a/subsubdir_c',
                     'root/subdir_b',
                     'root/subdir_c']

            .. group-tab:: Excluding intermediates

                .. code-block:: python

                    >>> get_subdirectories('root', include_intermediates=False)
                    ['root/subdir_a/subsubdir_b',
                     'root/subdir_a/subsubdir_c',
                     'root/subdir_b',
                     'root/subdir_c']
    """
    dirs = [root]
    subdirs = set()

    while len(dirs) > 0:
        _dirs = []
        for cdir in dirs:
            csubdirs = [j(cdir, d) for d in os.listdir(cdir) if os.path.isdir(j(cdir, d))]
            if len(csubdirs) == 0:
                subdirs.add(cdir)
            else:
                if include_intermediates:
                    subdirs.add(cdir)
                _dirs.extend(csubdirs)

        dirs = _dirs

    return subdirs


def match(root: str, pattern: str, sort_by: str = None) -> Dict[str, dict]:
    """
    Find and return information about subdirectories of a root that match a given pattern.

    Args:
        root: the root of the subdirectories to look in.
        pattern: a string specifying the pattern the subdirectories should correspond to.
            It should look similar to a format string, without the ``f`` in front of the string.
            Inside curly braces you can put a variable name, which you can later extract from the results.
            Anything inside curly braces will be matched to word characters (``[a-zA-Z0-9_-]``) including dashes and underscores.

        sort_by: the key to sort the results by. If not given, the results will be returned in the order they were found.

    Returns:
        A |Result| object containing the matched directories as keys and information (also |Result| object) about those matches as the values. Each information dictionary contains the variables given in the pattern.
        E.g. using a pattern such as ``{a}/{b}/{c}`` will populate the ``info.a``, ``info.b`` and ``info.c`` keys of the info |Result| object.

    Example:
        Given a file-structure as follows:

        .. code-block::

            root
            |- NH3-BH3
            |   |- BLYP_QZ4P
            |   |  |- extra_dir
            |   |  |- blablabla
            |   |
            |   |- BLYP_TZ2P
            |   |  |- another_dir
            |   |
            |   |- M06-2X_TZ2P
            |
            |- SN2
            |   |- BLYP_TZ2P
            |   |- M06-2X_TZ2P
            |   |  |- M06-2X_TZ2P

        We can run the following scripts to match the subdirectories.

        .. code-block:: python

            from tcutility import log
            # get the matches, we want to extract the system name (NH3-BH3 or SN2)
            # and the functional and basis-set
            # we don't want the subdirectories
            matches = match('root', '{system}/{functional}_{basis_set}')

            # print the matches as a table
            rows = []
            for d, info in matches.items():
                rows.append([d, info.system, info.functional, info.basis_set])

            log.table(rows, ['Directory', 'System', 'Functional', 'Basis-Set'])

        which prints

        .. code-block::

            [2024/01/17 14:39:08] Directory                  System    Functional   Basis-Set
            [2024/01/17 14:39:08] ───────────────────────────────────────────────────────────
            [2024/01/17 14:39:08] root/SN2/M06-2X_TZ2P       SN2       M06-2X       TZ2P
            [2024/01/17 14:39:08] root/NH3-BH3/BLYP_TZ2P     NH3-BH3   BLYP         TZ2P
            [2024/01/17 14:39:08] root/NH3-BH3/M06-2X_TZ2P   NH3-BH3   M06-2X       TZ2P
            [2024/01/17 14:39:08] root/SN2/BLYP_TZ2P         SN2       BLYP         TZ2P
            [2024/01/17 14:39:08] root/NH3-BH3/BLYP_QZ4P     NH3-BH3   BLYP         QZ4P

"""
    # get the number and names of substitutions in the given pattern
    substitutions = re.findall(r"{(\w+[+*?]?)}", pattern)
    # the pattern should resolve to words and may contain - and _
    # replace them here
    for sub in substitutions:
        quantifier = sub[-1] if sub[-1] in "+*?" else "+"
        pattern = pattern.replace("{" + sub + "}", f"([a-zA-Z0-9_-]{quantifier})")

    ret = results.Result()
    # root dir can be any level deep. We should count how many directories are in root
    root_length = len(split_all(root))
    # get all subdirectories first, we can loop through them later
    subdirs = get_subdirectories(root, include_intermediates=True)
    # remove the root from the subdirectories. We cannot use str.removeprefix because it was added in python 3.9
    subdirs = [j(*split_all(subdir)[root_length:]) for subdir in subdirs if len(split_all(subdir)[root_length:]) > 0]
    for subdir in subdirs:
        # check if we get a match with our pattern
        match = re.fullmatch(pattern, subdir)
        if not match:
            continue

        p = j(root, subdir)
        # get the group data and add it to the return dictionary. We skip the first group because it is the full directory path
        ret[p] = results.Result(directory=p, **{substitutions[i]: match.group(i + 1) for i in range(len(substitutions))})

    if not sort_by:
        return ret

    return results.Result(sorted(ret.items(), key=lambda d: d[1][sort_by]))


def path_depth(path: str) -> int:
    """
    Calculate the depth of a given path.
    """
    return len(split_all(path))




from tcutility import spell_check


def test_wagner_fischer1():
    dist = spell_check.wagner_fischer('kitten', 'sitting')
    assert dist == 3

def test_wagner_fischer2():
    dist = spell_check.wagner_fischer('kitten', 'kitten')
    assert dist == 0

def test_wagner_fischer3():
    dist = spell_check.wagner_fischer('kitten', 'Kitten')
    assert dist == 1

def test_wagner_fischer4():
    dist = spell_check.wagner_fischer('kitten', 'Kitten', case_missmatch_cost=.5)
    assert dist == .5

def test_wagner_fischer5():
    dist = spell_check.wagner_fischer('kitten', 'sitting', case_missmatch_cost=.5)
    assert dist == 3

def test_wagner_fischer6():
    dist = spell_check.wagner_fischer('kitten', 'Kitten', substitution_cost=.5)
    assert dist == 1

def test_wagner_fischer7():
    dist = spell_check.wagner_fischer('kitten', 'sittin', substitution_cost=.5)
    assert dist == 1

def test_wagner_fischer8():
    dist = spell_check.wagner_fischer('kitten', 'kittttten', insertion_cost=.5)
    assert dist == 1.5

def test_wagner_fischer9():
    dist = spell_check.wagner_fischer('', 'kitten')
    assert dist == 6

def test_get_closest1():
    closest = spell_check.get_closest('kitten', ['mitten', 'bitten', 'sitting'])
    assert closest == ['mitten', 'bitten']

def test_get_closest2():
    closest = spell_check.get_closest('kitten', ['mitten', 'bitten', 'sitting'], compare_func=spell_check.naive_recursive)
    assert closest == ['mitten', 'bitten']

def test_get_closest3():
    closest = spell_check.get_closest('kitten', ['mitten', 'bitten', 'sitting'], ignore_chars='ening')
    assert closest == ['mitten', 'bitten', 'sitting']

def test_get_closest4():
    closest = spell_check.get_closest('kitten', ['Kitten', 'kItten', 'kiTten', 'KItten', 'KITten'], ignore_case=False)
    assert closest == ['Kitten', 'kItten', 'kiTten']

def test_get_closest5():
    closest = spell_check.get_closest('kitten', ['Kitten', 'kItten', 'kiTten', 'KItten', 'KITten', 'KITTEN'], ignore_case=True)
    assert closest == []

def test_get_closest6():
    closest = spell_check.get_closest('kitten', ['Kitten', 'kItten', 'kiTten', 'KItten', 'KITten', 'KITTEN'], maximum_distance=2)
    assert closest == ['Kitten', 'kItten', 'kiTten', 'KItten']

if __name__ == '__main__':
    import pytest

    pytest.main()



import os
import re
from datetime import datetime
from typing import List

import numpy as np
from scm import plams

from tcutility import constants, ensure_list
from tcutility.results import Result, cache
from tcutility.tc_typing import arrays

j = os.path.join


def get_calc_files(calc_dir: str) -> dict:
    """Function that returns files relevant to AMS calculations stored in ``calc_dir``.

    Args:
        calc_dir: path pointing to the desired calculation

    Returns:
        Dictionary containing filenames and paths
    """
    # collect all files in the current directory and subdirectories
    files = []
    for root, _, files_ in os.walk(os.path.abspath(calc_dir)):
        if os.path.split(root)[1].startswith("."):
            continue

        # some results are stored in dirs called {name}.results, if the calculations uses fragments there will be additional dirs called
        # {name}.{fragname}.results, which do not contain new information as the required info is copied over to {name}.results. Therefore
        # we should skip the fragment directories
        if os.path.split(root)[1].endswith(".results") and len(os.path.split(root)[1].split(".")) > 2:
            continue

        # we only need the rkf files in the .results directories
        if root.endswith(".rkf") and ".results" not in root:
            continue

        # opened files end in ~ and should be ignored
        if root.endswith("~"):
            continue

        # t21 files for specific atoms are not usefull and should be ignored
        if "t21." in root:
            continue

        # if a calculation is running or did not finish correctly there might be tmp. dirs, which should be ignored
        if os.path.split(root)[1].startswith("tmp."):
            continue

        # the rest can all be stored
        files.extend([j(root, file) for file in files_])

    # parse the filenames
    ret = {}
    ret["root"] = os.path.abspath(calc_dir)
    for file in files:
        # rkf files are either called /{program}.rkf, or /{name}.{program}.rkf, so we must parse them in a special way
        if file.endswith(".rkf"):
            f = os.path.split(file)[1]
            parts = f.split(".")
            f = f"{parts[-2]}.rkf"
            ret[f] = os.path.abspath(file)

        # logfiles are generally called either {name}.log or {name}.logfile
        if file.endswith(".logfile") or file.endswith(".log") or file.endswith(".err"):
            f = os.path.split(file)[1]
            parts = f.split(".")
            f = ".".join(parts[1:]).replace("logfile", "log")
            ret[f] = os.path.abspath(file)

        if file.endswith(".out") and "CreateAtoms" not in file:
            ret["out"] = os.path.abspath(file)

    return ret


def get_ams_version(calc_dir: str) -> Result:
    """Function to get the AMS version used in the calculation.

    Args:
        calc_dir: path pointing to the desired calculation.

    Returns:
        :Dictionary containing results about the AMS version:

            - **full (str)** – the full version string as written by SCM.
            - **major (str)** – major AMS version, should correspond to the year of release.
            - **minor (str)** – minor AMS version.
            - **micro (str)** – micro AMS version, should correspond to the internal revision number.
            - **date (datetime.datetime)** – date the AMS version was released.
    """
    ret = Result()
    files = get_calc_files(calc_dir)
    reader_ams = cache.get(files["ams.rkf"])

    # store information about the version of AMS
    ret.full = str(reader_ams.read("General", "release"))
    # decompose the full version string
    ret.major = ret.full.split(".")[0]
    ret.minor = ret.full.split()[0].split(".")[1]
    ret.micro = ret.full.split()[1]
    ret.date = datetime.strptime(ret.full.split()[-1][1:-1], "%Y-%m-%d")

    return ret


def get_ams_info(calc_dir: str) -> Result:
    """Function to read useful info about the calculation in ``calc_dir``. Returned information will depend on the type of file that is provided.

    Args:
        calc_dir: path pointing to the desired calculation.

    Returns:
        :Dictionary containing results about the calculation and AMS:

            - **ams_version (Result)** – information about the AMS version used, see :func:`get_ams_version`.
            - **engine (str)** – the engine that was used to perform the calculation, for example 'adf', 'dftb', ...
            - **job_id (str)** – the ID of the job, can be used to check if two calculations are the same. Might also be used as a unique identifier for the calculation.
            - **status (Result)** – information about calculation status, see :func:`get_calculation_status`.
            - **is_multijob (bool)** – whether the job was a multijob, for example a fragment analysis.
            - **molecule (Result)** – information about the input and output molecules and the molecular system in general, see :func:`get_molecules`.
            - **history (Result)** – information about history variables, see :func:`get_history`.
    """
    ret = Result()
    ret.files = get_calc_files(calc_dir)
    reader_ams = cache.get(ret.files["ams.rkf"])

    # check what the program is first. The program can be either AMS or one of the engines (ADF, DFTB, ...)
    if ("General", "engine") in reader_ams:
        ret.engine = str(reader_ams.read("General", "engine")).strip().lower()
    # if program cannot be read from reader it is probably an old version of ADF, so we should default to ADF
    else:
        ret.engine = "adf"

    # store the input of the calculation
    ret.input = get_ams_input(reader_ams.read("General", "user input"))

    # store the job id, which should be unique for the job
    ret.job_id = reader_ams.read("General", "jobid") if ("General", "jobid") in reader_ams else None

    # store information about the version of AMS
    ret.ams_version = get_ams_version(calc_dir)

    # store the computation timings, only available in ams.rkf
    ret.timing = get_timing(calc_dir)

    # store the calculation status
    ret.status = get_calculation_status(calc_dir)

    # check if this was a multijob
    ret.is_multijob = False
    if len([file for file in ret.files if file.endswith(".rkf")]) > 2:
        ret.is_multijob = True

    # read molecules
    ret.molecule = get_molecules(calc_dir)

    # and history variables
    ret.history = get_history(calc_dir)

    ret.pes = get_pes(calc_dir)

    cache.unload(ret.files["ams.rkf"])
    return ret


def get_timing(calc_dir: str) -> Result:
    """Function to get the timings from the calculation.

    Args:
        calc_dir: path pointing to the desired calculation.

    Returns:
        :Dictionary containing results about the timings:

            - **cpu (float)** – time spent performing calculations on the cpu.
            - **sys (float)** – time spent by the system (file IO, process creation/destruction, etc ...).
            - **total (float)** – total time spent by AMS on the calculation, can be larger than the sum of cpu and sys.
    """
    ret = Result()
    files = get_calc_files(calc_dir)
    reader_ams = cache.get(files["ams.rkf"])

    ret.cpu = reader_ams.read("General", "CPUTime") if ("General", "CPUTime") in reader_ams else None
    ret.sys = reader_ams.read("General", "SysTime") if ("General", "SysTime") in reader_ams else None
    ret.total = reader_ams.read("General", "ElapsedTime") if ("General", "ElapsedTime") in reader_ams else None

    return ret


def get_calculation_status(calc_dir: str) -> Result:
    """Function that returns the status of the calculation described in reader.
    In case of non-succes it will also give possible reasons for the errors/warnings.

    Args:
        reader: ``plams.KFReader`` or ``plams.KFFile`` object pointing to the desired calculation

    Returns:
        :Dictionary containing information about the calculation status:

            - **fatal (bool)** – True if calculation cannot be analysed correctly, False otherwise
            - **reasons (list[str])** – list of reasons to explain the status, they can be errors, warnings, etc.
            - **name (str)** – calculation status written as a string, one of ("SUCCESS", "RUNNING", "UNKNOWN", "SUCCESS(W)", "FAILED")
            - **code (str)** – calculation status written as a single character, one of ("S", "R", "U", "W" "F")
    """
    files = get_calc_files(calc_dir)
    reader_ams = cache.get(files["ams.rkf"])

    ret = Result()
    ret.fatal = True
    ret.name = None
    ret.code = None
    ret.reasons = []

    termination_status = str(reader_ams.read("General", "termination status")).strip()

    # parse the logfile to find errors and warnings
    if "log" in files:
        with open(files["log"]) as logfile:
            for line in logfile.readlines():
                # the first 25 characters include the timestamp and two spaces
                line_ = line.strip()[25:]
                # errors and warnings have a predictable format
                if line_.lower().startswith("error:") or line_.lower().startswith("warning: "):
                    ret.reasons.append(line_)

    if termination_status == "NORMAL TERMINATION":
        ret.fatal = False
        ret.name = "SUCCESS"
        ret.code = "S"
        return ret

    if termination_status == "IN PROGRESS":
        ret.fatal = False
        ret.reasons.append("Calculation in progress")
        ret.name = "RUNNING"
        ret.code = "R"
        return ret

    if termination_status == "UNKNOWN":
        ret.reasons.append("Calculation status unknown")
        ret.name = "UNKNOWN"
        ret.code = "U"
        return ret

    if termination_status == "NORMAL TERMINATION with warnings":
        ret.fatal = False
        ret.name = "SUCCESS(W)"
        ret.code = "W"
        return ret

    if termination_status == "NORMAL TERMINATION with errors":
        ret.name = "FAILED"
        ret.code = "F"
        return ret

    # if we have not excited the function yet we do not know what the status is
    # probably means that there was a parsing error in ams, which will be placed in termination status
    ret.reasons.append(termination_status)
    ret.name = "UNKNOWN"
    ret.code = "U"
    return ret


# ------------------------------------------------------------- #
# ------------------------- Geometry -------------------------- #
# ------------------------------------------------------------- #


# -------------------- Fragment indices ----------------------- #


def _get_fragment_indices_from_input_order(results_type) -> arrays.Array1D:
    """Function to get the fragment indices from the input order. This is needed because the fragment indices are stored in internal order in the rkf file."""
    frag_indices = np.array(results_type.read("Geometry", "fragment and atomtype index")).reshape(2, -1)  # 1st row: Fragment index; 2nd row atomtype index
    atom_order = np.array(results_type.read("Geometry", "atom order index")).reshape(2, -1)  # 1st row: input order; 2nd row: internal order
    return frag_indices[0][atom_order[0] - 1]


# ------------------------ Molecule -------------------------- #


def _make_molecule(kf_variable: str, reader_ams: plams.KFReader, natoms: int, atnums: List[int]) -> plams.Molecule:
    """Makes a plams.Molecule object from the given kf_variable ("InputMolecule" or "Molecule") and reader."""
    # read output molecule
    ret_mol = plams.Molecule()
    coords = np.array(reader_ams.read(kf_variable, "Coords")).reshape(natoms, 3) * constants.BOHR2ANG
    for atnum, coord in zip(atnums, coords):
        ret_mol.add_atom(plams.Atom(atnum=atnum, coords=coord))
    if ("Molecule", "fromAtoms") in reader_ams and ("Molecule", "toAtoms") in reader_ams and ("Molecule", "bondOrders") in reader_ams:
        at_from = ensure_list(reader_ams.read("InputMolecule", "fromAtoms"))
        at_to = ensure_list(reader_ams.read("InputMolecule", "toAtoms"))
        bos = ensure_list(reader_ams.read("InputMolecule", "bondOrders"))
        for at1, at2, order in zip(at_from, at_to, bos):
            ret_mol.add_bond(plams.Bond(ret_mol[at1], ret_mol[at2], order=order))
    else:
        ret_mol.guess_bonds()
    return ret_mol


def get_molecules(calc_dir: str) -> Result:
    """
    Function to get molecules from the calculation, including input, output and history molecules.
    It will also add bonds to the molecule if they are given in the rkf file, else it will guess them.

    Args:
        calc_dir: path pointing to the desired calculation.

    Returns:
        :Dictionary containing information about the molecular systems:

            - **number_of_atoms (int)** – number of atoms in the molecule.
            - **atom_numbers (list[int])** – list of atomic numbers for each atom in the molecule.
            - **atom_symbols (list[str])** – list of elements for each atom in the molecule.
            - **atom_masses (list[float])** – list of atomic masses for each atom in the molecule.
            - **input (plams.Molecule)** – molecule that was given in the input for the calculation.
            - **output (plams.Molecule)** – final molecule that was given as the output for the calculation. If the calculation was a singlepoint calculation output and input molecules will be the same.
            - **frag_indices (numpy array[int] (1D))** – list of fragment indices for each atom in the molecule. The indices are given in the order of the atoms in the molecule.
    """
    files = get_calc_files(calc_dir)
    # all info is stored in reader_ams
    reader_ams = cache.get(files["ams.rkf"])

    ret = Result()

    # read general
    atnums = ensure_list(reader_ams.read("InputMolecule", "AtomicNumbers"))  # type: ignore plams does not include type hints. Returns list[int]
    natoms = len(atnums)
    ret.number_of_atoms = natoms
    ret.atom_numbers = atnums
    ret.atom_symbols = str(reader_ams.read("InputMolecule", "AtomSymbols")).split()
    ret.atom_masses = reader_ams.read("InputMolecule", "AtomMasses")

    # read input molecule
    ret.input = _make_molecule("InputMolecule", reader_ams, natoms, atnums)

    # read output molecule
    ret.output = _make_molecule("Molecule", reader_ams, natoms, atnums)

    # add fragment indices to the molecule using adf file
    try:
        reader_adf = cache.get(files["adf.rkf"])
        ret.frag_indices = _get_fragment_indices_from_input_order(reader_adf)
    except KeyError:
        ret.frag_indices = np.ones(natoms)

    # add the charges to the molecule
    try:
        ret.mol_charge = reader_ams.read("Molecule", "Charge")
    except KeyError:
        ret.mol_charge = 0.0

    return ret


def get_pes(calc_dir: str) -> Result:
    """
    Function to get PES scan variables.

    Args:
        calc_dir: path pointing to the desired calculation.

    Returns:
        :Dictionary containing information about the PES scan:

            - ``nscan_coords`` **int** – the number of scan-coordinates for this PES scan.
            - ``scan_coord_name`` **list[str] | str** – the names of the scan-coordinates.
                If there is more than one scan-coordinates it will be a list of strings, otherwise it will be a single string.
            - ``scan_coord`` **list[np.ndarray] | np.ndarray** – arrays of values for the scan-coordinates.
                If there is more than one scan-coordinates it will be a list of arrays, otherwise it will be a single array.
            - ``npoints`` **list[int] | int** – number of scan points for the scan-coordinates.
                If there is more than one scan-coordinates it will be a list of integers, otherwise it will be a single integer.
    """
    # read history mols
    files = get_calc_files(calc_dir)
    # all info is stored in reader_ams
    reader_ams = cache.get(files["ams.rkf"])

    # check if we have done a PESScan
    if ("PESScan", "nPoints") not in reader_ams:
        return

    ret = Result()

    ret.nscan_coords = reader_ams.read("PESScan", "nScanCoord")
    ret.scan_coord_name = [reader_ams.read("PESScan", f"ScanCoord({i+1})").strip() for i in range(ret.nscan_coords)]
    ret.npoints = [reader_ams.read("PESScan", f"nPoints({i+1})") for i in range(ret.nscan_coords)]
    ret.scan_coord = [np.linspace(reader_ams.read("PESScan", f"RangeStart({i+1})"), reader_ams.read("PESScan", f"RangeEnd({i+1})"), ret.npoints[i]) for i in range(ret.nscan_coords)]
    if ("PESScan", "PES") in reader_ams:
        ret.energies = np.array(reader_ams.read("PESScan", "PES")).reshape(*ret.npoints) * constants.HA2KCALMOL

    if ret.nscan_coords == 1:
        ret.scan_coord_name = ret.scan_coord_name[0]
        ret.scan_coord = ret.scan_coord[0]
        ret.npoints = ret.npoints[0]

    return ret


def get_history(calc_dir: str) -> Result:
    """
    Function to get history variables. The type of variables read depends on the type of calculation.

    Args:
        calc_dir: path pointing to the desired calculation.

    Returns:
        :Dictionary containing information about the calculation status:

            - ``number_of_entries`` **(int)** – number of steps in the history.
            - ``{variable}`` **(list[Any])** – variable read from the history section. The number of variables and type of variables depend on the nature of the calculation.

              Common variables:
                | ``Molecule`` **(list[plams.Molecule])** – list of molecules from the history, for example from a geometry optimization or PES scan.
                | ``energy`` **(list[float])** – list of energies associated with each geometry step.
                | ``gradient`` **(list[list[float]])** – array of gradients for each geometry step and each atom.
    """
    # read history mols
    files = get_calc_files(calc_dir)
    # all info is stored in reader_ams
    reader_ams = cache.get(files["ams.rkf"])

    ret = Result()

    # read general info
    atnums = ensure_list(reader_ams.read("InputMolecule", "AtomicNumbers"))  # type: ignore plams does not include type hints. Returns list[int]
    natoms = len(atnums)

    if ("History", "nEntries") in reader_ams:
        # the number of elements per history variable (e.g. number of steps in a geometry optimization)
        ret.number_of_entries = reader_ams.read("History", "nEntries")
        # for history we read the other variables in the rkf file
        # the variables are given in ('History', f'ItemName({index})') so we have to find all of those
        index = 1  # tracks the current index of the variable. Will be set to False to stop the search
        items = []
        while index:
            # try to access ItemName, if it cannot find it stop iteration
            try:
                item_name = reader_ams.read("History", f"ItemName({index})")
                items.append(item_name)
                index += 1
            except KeyError:
                index = False

        if "Coords" in items:
            items.append("Molecule")

        # create variable in result for each variable found
        for item in items:
            ret[item.lower()] = []

        # collect the elements for each history variable
        for i in range(ret.number_of_entries):
            for item in items:
                # Molecule are special, because we will convert them to plams.Molecule objects first
                if item == "Molecule":
                    mol = plams.Molecule()
                    coords = np.array(reader_ams.read("History", f"Coords({i+1})")).reshape(natoms, 3) * 0.529177
                    for atnum, coord in zip(atnums, coords):
                        mol.add_atom(plams.Atom(atnum=atnum, coords=coord))
                    mol.guess_bonds()
                    ret.molecule.append(mol)
                # other variables are just added as-is
                else:
                    val = reader_ams.read("History", f"{item}({i+1})")
                    ret[item.lower()].append(val)

        if "converged" not in ret.keys() and ("PESScan", "HistoryIndices") in reader_ams:
            ret["converged"] = [False] * ret.number_of_entries
            for idx in ensure_list(reader_ams.read("PESScan", "HistoryIndices")):
                ret["converged"][idx - 1] = True

    return ret


def get_input_blocks():
    """
    This function reads input_blocks and decomposes its content into a list of blocks and a list of non-standard blocks
    The general format is as follows:

    parentblock
    - subblock
    - - subsubblock
    - subblock !nonstandard
    parentblock
    - subblock
    - - subsubblock
    - - - subsubsubblock
    - - - subsubsubblock !nonstandard

    Each subblock has to be defined within its parent block. !nonstandard indicates that the block is a non-standard block
    These blocks are special in that they can contain multiple of the same entry
    """
    blocks = []
    nonstandard_blocks = []
    parent_blocks = []  # this list tracks the parent blocks of the current block
    with open(j(os.path.split(__file__)[0], "input_blocks")) as inpblx:
        lines = inpblx.readlines()

    for line in lines:
        line = line.strip().lower()
        # we can ignore some lines
        if line == "" or line.startswith("#"):
            continue

        # block_depth indicates how many parents the block has
        block_depth = line.count("- ")
        # we reduce the amount of parent_blocks using block_depth
        # if we move from a subsubblock to a subblock we remove the last-added block
        parent_blocks = parent_blocks[:block_depth]

        # remove the "- " substrings
        block = line.split("- ")[-1].strip().lower()
        # check if the block is non-standard. If it is, remove the !nonstandard substring
        # and add it to the nonstandard_blocks list
        if block.endswith("!nonstandard"):
            block = block.split()[0]
            nonstandard_blocks.append(parent_blocks.copy() + [block])
        # in both standard and nonstandard cases add the block to blocks list
        blocks.append(parent_blocks.copy() + [block])
        # add the current block to parent_blocks for the next line
        parent_blocks.append(block.lower())

    return blocks, nonstandard_blocks


def get_ams_input(inp: str) -> Result:
    def get_possible_blocks():
        # check which blocks are openable given the current open blocks
        # we start by considering all blocks
        possible_blocks = blocks
        # we iterate through the open blocks and check if the first element in the possible_blocks is the same.
        # this way we move down through the blocks
        for open_block in open_blocks:
            possible_blocks = [block[1:] for block in possible_blocks if len(block) > 0 and block[0].lower() == open_block.lower()]
        # we want only the first element in the possible blocks, not the tails
        possible_blocks = set([block[0] for block in possible_blocks if len(block) > 0])
        return possible_blocks

    sett = Result()

    blocks, nonstandard_blocks = get_input_blocks()

    open_blocks = ["ams"]
    for line in inp.splitlines():
        line = line.strip()

        # we remove comments from the line
        # comments can be either at the start of a line or after a real statement
        # so we have to search the line for comment starters and remove the part after it
        for comment_start in ["#", "!", "::"]:
            if comment_start in line:
                idx = line.index(comment_start)
                line = line[:idx]

        # skip empty lines
        if not line:
            continue

        # if we encounter an end statement we can close the last openblock
        if line.lower().startswith("end"):
            open_blocks.pop(-1)
            continue

        # check if we are opening a block
        # We have to store the parts of a compact block (if it is compact)
        # it will be a list of tuples of key-value pairs
        compact_parts = None
        skip = False
        # check if the current line corresponds to a possible block
        for possible_block in get_possible_blocks():
            if line.lower().startswith(possible_block.lower()):
                # a block opening can either span multiple lines or can be use with compact notation
                # compact notation will always have = signs
                if "=" in line:
                    # split the key-values using some regex magic
                    compact = line[len(possible_block) :].strip()  # remove the block name
                    compact_parts = re.findall(r"""(\S+)=['"]{0,1}([^'"\n]*)['"]{0,1}""", compact)
                else:
                    skip = True
                # add the new block to the open_blocks
                open_blocks.append(possible_block)

        # if we are not in a compact block we can just skip
        if skip:
            continue

        # get the sett to the correct level
        # first check if the block is nonstandard
        is_nonstandard = [block.lower() for block in open_blocks] in nonstandard_blocks
        sett_ = sett
        # go to the layer one above the lowest
        for open_block in open_blocks[:-1]:
            sett_ = sett_[open_block]
        # at the lowest level we have to check if the block is nonstandard. If it is, we add the whole line to the settings object
        if is_nonstandard and not sett_[open_blocks[-1]]:
            sett_[open_blocks[-1]] = []
        # then finally go to the lowest sett layer
        sett_ = sett_[open_blocks[-1]]

        # if we are in a compact block we just add all the key-value pairs
        if compact_parts:
            for key, val in compact_parts:
                sett_[key] = val
            # compact blocks do not have end statements, so we can remove it again from the open_blocks
            open_blocks.pop(-1)
        # in a normal block we split the line and add them to the settings
        else:
            # for nonstandard blocks we add the line to the list
            if is_nonstandard:
                sett_.append(line.strip())
            # for normal blocks we split the line and set it as a key, the rest of the line is the value
            else:
                sett_[line.split()[0]] = line[len(line.split()[0]) :].strip()

    for engine_block in ["engine adf", "engine dftb", "engine band"]:
        if engine_block not in sett["ams"]:
            continue
        sett["ams"][engine_block.split()[1]] = sett["ams"][engine_block]
        del sett["ams"][engine_block]

    return sett["ams"]



import os
import tempfile
import pytest
from tcutility import pathfunc


@pytest.fixture
def temp_dir():
    '''
    This will create a directory structure like:

    root
    |- subdir_a
    |  |- subsubdir_b
    |  |- subsubdir_c
    |- subdir_b
    |- subdir_c
    '''
    with tempfile.TemporaryDirectory() as temp_dir:
        # Create directories
        os.makedirs(os.path.join(temp_dir, "subdir_a", "subsubdir_b"))
        os.makedirs(os.path.join(temp_dir, "subdir_a", "subsubdir_c"))
        os.makedirs(os.path.join(temp_dir, "subdir_b"))
        os.makedirs(os.path.join(temp_dir, "subdir_c"))
        yield temp_dir

@pytest.fixture
def temp_dir2():
    '''
    This will create a directory structure like:

    root
    |- NH3-BH3
    |   |- BLYP_QZ4P
    |   |  |- extra_dir
    |   |  |- blablabla
    |   |
    |   |- BLYP_TZ2P
    |   |  |- another_dir
    |   |
    |   |- M06-2X_TZ2P
    |
    |- SN2
    |   |- BLYP_TZ2P
    |   |- M06-2X_TZ2P
    |   |  |- M06-2X_TZ2P
    '''
    with tempfile.TemporaryDirectory() as temp_dir:
        # Create directories
        os.makedirs(os.path.join(temp_dir, "NH3-BH3", "BLYP_QZ4P", "extra_dir"))
        os.makedirs(os.path.join(temp_dir, "NH3-BH3", "BLYP_QZ4P", "blablabla"))
        os.makedirs(os.path.join(temp_dir, "NH3-BH3", "BLYP_TZ2P", "another_dir"))
        os.makedirs(os.path.join(temp_dir, "NH3-BH3", "M06-2X_TZ2P"))
        os.makedirs(os.path.join(temp_dir, "SN2", "BLYP_TZ2P"))
        os.makedirs(os.path.join(temp_dir, "SN2", "M06-2X_TZ2P", "M06-2X_TZ2P"))
        yield temp_dir


def test_get_subdirectories(temp_dir):
    expected = {
        f'{temp_dir}/subdir_a/subsubdir_b',
        f'{temp_dir}/subdir_a/subsubdir_c',
        f'{temp_dir}/subdir_b',
        f'{temp_dir}/subdir_c'
    }
    assert pathfunc.get_subdirectories(temp_dir) == expected


def test_get_subdirectories2(temp_dir):
    expected = {
        f'{temp_dir}',
        f'{temp_dir}/subdir_a',
        f'{temp_dir}/subdir_a/subsubdir_b',
        f'{temp_dir}/subdir_a/subsubdir_c',
        f'{temp_dir}/subdir_b',
        f'{temp_dir}/subdir_c'
    }
    assert pathfunc.get_subdirectories(temp_dir, include_intermediates=True) == expected


def test_match(temp_dir2):
    assert len(pathfunc.match(temp_dir2, '{system}/{functional}_{basis_set}')) == 5


def test_match2(temp_dir2):
    matches = pathfunc.match(temp_dir2, '{system}/{functional}_{basis_set}')
    assert f"{temp_dir2}/SN2/M06-2X_TZ2P" in matches
    assert matches[f"{temp_dir2}/SN2/M06-2X_TZ2P"].system == 'SN2'
    assert matches[f"{temp_dir2}/SN2/M06-2X_TZ2P"].functional == 'M06-2X'
    assert matches[f"{temp_dir2}/SN2/M06-2X_TZ2P"].basis_set == 'TZ2P'


def test_split_all():
    assert pathfunc.split_all("a/b/c") == ["a", "b", "c"]


def test_split_all2():
    assert pathfunc.split_all("a") == ["a"]


def test_split_all3():
    # empty string corresponds to the `.` directory, i.e. here
    assert pathfunc.split_all("") == ["."]


def test_split_all4():
    import os

    assert pathfunc.split_all(os.path.join("a", "b", "c")) == ["a", "b", "c"]


def test_path_depth():
    assert pathfunc.path_depth('a/b/c') == 3


def test_path_depth2():
    assert pathfunc.path_depth('a/One Drive/c') == 3
    
def test_path_depth4():
    assert pathfunc.path_depth('a') == 1
    


if __name__ == "__main__":
    import pytest

    pytest.main(['-vv'])



import requests
import json
import os
from tcutility import spell_check, cache



# @cache.cache_file('dois')
@cache.cache
def _get_doi_data(doi: str) -> dict:
	'''
	Get information about an article using the crossref.org API.

	Args:
		doi: the DOI to get information about.
	'''
	print(f'http://api.crossref.org/works/{doi}')
	data = requests.get(f'http://api.crossref.org/works/{doi}').text
	if data == 'Resource not found.':
		raise ValueError(f'Could not find DOI {doi}.')
	data = json.loads(data)
	return data


# @cache.cache_file('journal_abbrvs')
@cache.cache
def _get_journal_abbreviation(journal: str) -> str:
	'''
	Get the journal name abbreviation using the abbreviso API.

	Args:
		journal: the name of the journal to get the abbreviation of.
	'''
	return requests.get(f"https://abbreviso.toolforge.org/a/{journal}").text


@cache.cache
def _get_publisher_city(publisher: str) -> str:
	'''
	Get the city of a publisher.
	'''
	with open(os.path.join(os.path.split(__file__)[0], 'data', 'cite', '_publisher_cities.json')) as cities:
		cities = json.loads(cities.read())
	return cities.get(publisher)


def cite(doi: str, style: str = 'wiley', mode='html') -> str:
	'''
	Format an article in a certain style.

	Args:
		doi: the article DOI to generate a citation for.
		style: the style formatting to use. Can be ``['wiley', 'acs', 'rsc']``.
		mode: the formatting mode. Can be ``['html', 'latex', 'plain']``.
	'''
	# check if the style was correctly given
	spell_check.check(style, ['wiley', 'acs', 'rsc'])
	spell_check.check(mode, ['html', 'latex', 'plain'])

	# get the information about the DOI
	data = _get_doi_data(doi)

	if data['message']['type'] == 'journal-article':
		citation = _format_article(data, style)

	if data['message']['type'] == 'book-chapter' or (data['message']['type'] == 'other' and 'ISBN' in data['message']):
		citation = _format_book_chapter(data, style)

	if mode == 'plain':
		citation = citation.replace('<i>', '')
		citation = citation.replace('</i>', '')
		citation = citation.replace('<b>', '')
		citation = citation.replace('</b>', '')

	if mode == 'latex':
		citation = citation.replace('<i>', r'\textit{')
		citation = citation.replace('</i>', '}')
		citation = citation.replace('<b>', r'\textbf{')
		citation = citation.replace('</b>', '}')

	return citation


def get_pages(data):
	try:
		pages = data['message'].get('page').replace('-', '–')
		return pages
	except AttributeError:
		pages = '???'

	url = data['message']['URL']
	if 'ceur.' in url:
		return 'e' + url.split('ceur.')[-1]


def is_accepted(data):
	for assertion in data['message'].get('assertion', []):
		if assertion['name'] == 'accepted':
			return True
	return False


def _format_article(data: dict, style: str) -> str:
	# grab usefull data
	journal = data['message']['container-title'][0]
	journal_abbreviation = _get_journal_abbreviation(journal)
	year = data['message']['issued']['date-parts'][0][0]
	volume = data['message'].get('volume', '')
	pages = get_pages(data)
	title = data['message']['title'][0]
	doi = data['message']['DOI']

	accepted = is_accepted(data)

	# Get the initials from the author given names
	# also store the family names
	initials = []
	last_names = []
	for author in data['message']['author']:
		# we get the capital letters from the first names
		# these will become the initials for this author
		firsts = [char + '.' for char in author['given'].title() if char.isupper()]
		firsts = " ".join(firsts)
		initials.append(firsts)
		last_names.append(author['family'].title())

	# format the citation correctly
	if style == 'wiley':
		names = [f'{first} {last}' for first, last in zip(initials, last_names)]
		citation = f'{", ".join(names)}, <i>{journal_abbreviation}</i> <b>{year}</b>, <i>{volume}</i>'
		if pages:
			citation += f', {pages}'
		citation += '.'

	elif style == 'acs':
		names = [f'{last}, {first}' for first, last in zip(initials, last_names)]
		citation = f'{"; ".join(names)} {title} <i>{journal_abbreviation}</i> <b>{year}</b>, <i>{volume}</i>'
		if pages:
			citation += f', {pages}'
		citation += f'. DOI: {doi}'

	elif style == 'rsc':
		names = [f'{first} {last}' for first, last in zip(initials, last_names)]
		citation = f'{", ".join(names)}, <i>{journal_abbreviation}</i> {year}, <b>{volume}</b>'
		if pages:
			citation += f', {pages}'
		citation += '.'

	return citation


def _format_book_chapter(data: dict, style: str) -> str:
	# grab usefull data
	publisher = data['message']['publisher']
	year = data['message']['published-print']['date-parts'][0][0]
	pages = data['message'].get('page')
	book_title = data['message']['container-title'][0]
	chapter_title = data['message']['title'][0]
	city = _get_publisher_city(publisher)

	for isbn in data['message']['isbn-type']:
		if isbn['type'] == 'electronic':
			original_book_data = _get_doi_data(f"{data['message']['prefix']}/{isbn['value']}")
			break

	# Get the initials from the author given names
	# also store the family names
	n_authors = len(data['message']['author'])
	initials = []
	last_names = []
	for author in data['message']['author']:
		# we get the capital letters from the first names
		# these will become the initials for this author
		firsts = [char + '.' for char in author['given'].title() if char.isupper()]
		firsts = " ".join(firsts)
		initials.append(firsts)
		last_names.append(author['family'].title())

	n_editors = len(original_book_data['message']['editor'])
	editor_initials = []
	editor_last_names = []
	for author in original_book_data['message']['editor']:
		# we get the capital letters from the first names
		# these will become the initials for this author
		firsts = [char + '.' for char in author['given'].title() if char.isupper()]
		firsts = " ".join(firsts)
		editor_initials.append(firsts)
		editor_last_names.append(author['family'].title())

	# format the citation correctly
	if style == 'wiley':
		names = [f'{last}, {first}' for first, last in zip(initials, last_names)]
		editors = [f'{first} {last}' for first, last in zip(editor_initials, editor_last_names)]
		if n_authors == 1:
			names = names[0]

		if 1 < n_authors < 4:
			names = ", ".join(names[:-1]) + ' and ' + names[-1]

		if n_authors >= 4:
			names = ", ".join(names[:3]) + ' et al.'

		if n_editors == 1:
			editors = editors[0]

		if 1 < n_editors < 4:
			editors = ", ".join(editors[:-1]) + ' and ' + editors[-1]

		if n_editors >= 4:
			editors = ", ".join(editors[:3]) + ' et al.'

		citation = f'{names} ({year}). {chapter_title}. In: <i>{book_title}</i> (ed. {editors}), {pages}. {city}: {publisher}'

	elif style == 'acs':
		raise NotImplementedError('No support for ACS style yet')

	elif style == 'rsc':
		raise NotImplementedError('No support for RSC style yet')

	return citation





from tcutility.data import basis_sets
import pytest


def test_number_of_orbitals():
    assert basis_sets.number_of_orbitals("Pd", "DZ") == 57


def test_number_of_virtuals():
    assert basis_sets.number_of_virtuals("Pd", "DZ") == 34


if __name__ == "__main__":
    pytest.main()



import os

from tcutility import cache, log


def _load_data() -> dict:
    """
    Load character width data. It is stored in a file where each line is formatted as:

            {font} {character} {width}

    Font names that originally contained spaces, for example "Times New Roman", will have the spaces replaced
    by underscores.

    Returns:
            We return a nested dictionary with the first key being the font and the nested key the characters.
            The widths are given in pixel sizes and are for font-size 1.
    """
    with open(os.path.join(os.path.split(__file__)[0], "_char_widths.txt"), encoding="utf-8") as widths:
        lines = widths.readlines()

    ret = {}
    for line in lines:
        # we split each line into parts by white space
        parts = line.split()
        # if there are only 2 parts, that means that the character was a space
        if len(parts) == 2:
            font, width = parts
            character = " "
        else:
            font, character, width = line.split()

        # fonts that contain spaces in the name had the spaces replaced by underscores
        # here we restore the original spaces
        font = font.replace("_", " ")
        width = float(width)

        ret.setdefault(font, {})
        ret[font][character] = width

    return ret


widths = _load_data()


@cache.cache
def get_pixel_width(character: str, font: str = "calibri", font_size: float = 11) -> float:
    """
    Get the pixel width of a character for a given font and font-size.
    For this we need to interpolate the pixel width between two reference values.
    In the _char_widths.txt file you will find for each font a set of character pixel widths for font-size 1.
    We then simply multiply the 1-width value by the desired value to get the correct pixel width.
    See

    Args:
            character: the character to get the pixel width for.
            font: the font family that is used.
            font_size: the font-size to get the pixel width at.
    """
    if font not in widths.keys():
        raise ValueError(f"Could not find font {font}.")

    return widths[font][character] * font_size


def text_width(text: str, font: str = "calibri", font_size: float = 11, mode: str = "pixels") -> float:
    """
    Get the width of a string for a given font and font size.

    Args:
            text: the object to determine the length for.
                    Will be cast to a string before calculating width.
            font: the font family that is used.
            font_size: the font-size to get the pixel width at.
            mode: the width-mode. Can be 'excel' or 'pixels'.

    Returns:
            The width of the text for a certain font and font size.
            If mode is 'excel' we pad it with 8 pixels and then divide by 7.6 (i.e. standard excel char width).
    """
    if font not in widths.keys():
        log.error(f"Could not find font {font}. Defaulting to Calibri.")
        font = "calibri"

    pixels = sum(get_pixel_width(char, font, font_size) for char in str(text))
    if mode == "excel":
        return (pixels + 8) / 7.6
    if mode == "word":
        return (pixels + 8) / 7.6

    return pixels


if __name__ == "__main__":
    print(text_width("this is a test text", font="test"))



import tcutility
import os
import numpy as np
import matplotlib.pyplot as plt

def read(path):
    return PyFragResult(path)


class PyFragResult:
    def __init__(self, path):
        self.path = path

        self._frag_results = {}
        self._step_results = []
        self._order = []
        self._mask = []
        self._properties = {}

        self._load()

    def _load(self):
        for frag_path, info in tcutility.pathfunc.match(self.path, 'frag_{frag}').items():
            self._frag_results[info.frag] = tcutility.results.read(frag_path)

        for step_path, info in tcutility.pathfunc.match(self.path, 'Step.{step}', sort_by='step').items():
            self._step_results.append({})
            for dir_name in os.listdir(step_path):
                p = os.path.join(step_path, dir_name)
                res = tcutility.results.read(p)
                self._step_results[-1][dir_name] = res

            self._order.append(int(info.step.removeprefix('0')))
            self._mask.append(True)

        self._mask = np.array(self._mask)

    def get_property(self, key: str, calc: str = 'complex'):
        if key in self._properties:
            return self._properties[key][self._order][self._mask[self._order]]
        p = np.array([res[calc].properties.get_multi_key(key) for res in self._step_results])
        return p[self._order][self._mask[self._order]]

    @property
    def fragments(self):
        return list(self._frag_results.keys())

    def __len__(self):
        return sum(self._mask)

    def get_geometry(self, *args, **kwargs):
        calc = kwargs.pop('calc', 'complex')
        g = np.array([tcutility.geometry.parameter(res[calc].molecule.input, *args, **kwargs) for res in self._step_results])
        return g[self._order][self._mask[self._order]]

    def sort_by(self, val: str or list, calc: str = 'complex'):
        if isinstance(val, str):
            val = self.get_property(val, calc=calc)
        self._order = np.argsort(val)

    def set_mask(self, mask: list):
        self._mask = mask

    def set_coord(self, coord: list):
        self._coord = coord

    def coord(self):
        return self._coord[self._order][self._mask[self._order]]

    def set_property(self, name: str, vals):
        self._properties[name] = vals

    @property
    def ts_idx(self):
        energies = self.total_energy()
        for i in range(len(self) - 2, 1, -1):
            if energies[i-1] < energies[i] and energies[i] < energies[i+1]:
                return i + 1

    def total_energy(self):
        return self.interaction_energy() + self.strain_energy()

    def interaction_energy(self):
        return self.get_property('energy.bond', 'complex')

    def strain_energy(self, fragment: str = None):
        if fragment is None:
            t = 0
            for frag in self.fragments:
                t += self.strain_energy(frag)
            return t

        return self.get_property('energy.bond', f'frag_{fragment}') - self._frag_results[fragment].properties.energy.bond

    def overlap(self, orb1, orb2, calc: str = 'complex', absolute: bool = True):
        S = np.array([orb.sfos[orb1] @ orb.sfos[orb2] for orb in self.orbs(calc)])
        if absolute:
            S = abs(S)
        return S[self._order][self._mask[self._order]]

    def orbital_energy_gap(self, orb1, orb2, calc: str = 'complex'):
        dE = np.array([abs(orb.sfos[orb1].energy - orb.sfos[orb2].energy) for orb in self.orbs(calc)])
        return dE[self._order][self._mask[self._order]]
    def sfo_coefficient(self, sfo, mo, calc: str = 'complex'):
        C = np.array([orb.sfos[sfo].coefficient(orb.mos[mo]) for orb in self.orbs(calc)])
        return C[self._order][self._mask[self._order]]

    @tcutility.cache.cache
    def orbs(self, calc: str = 'complex'):
        try:
            import pyfmo
        except ModuleNotFoundError:
            raise ModuleNotFoundError('The pyfmo module could not be loaded. To gain access please contact the TCutility developers!')


        return [pyfmo.orbitals2.objects.Orbitals(res[calc].files['adf.rkf']) for res in self._step_results]


if __name__ == '__main__':
    res_C = read('/Users/yumanhordijk/PhD/Projects/RadicalAdditionASMEDA/data/DFT/TS_C_O/PyFrag_OLYP_TZ2P')
    res_X = read('/Users/yumanhordijk/PhD/Projects/RadicalAdditionASMEDA/data/DFT/TS_X_O/PyFrag_OLYP_TZ2P')
    
    res_C.set_coord(res_C.get_geometry(0, 1))
    res_X.set_coord(res_X.get_geometry(1, 2))

    res_C.sort_by(res_C.coord())
    res_X.sort_by(res_X.coord())

    plt.figure()
    plt.plot(res_C.coord(), res_C.overlap('Methyl(SOMO)_A', 'Substrate(LUMO)_A'), label='C-addition')
    plt.plot(res_X.coord(), res_X.overlap('Methyl(SOMO)_A', 'Substrate(LUMO)_A'), label='X-addition')
    plt.ylabel(r'$\langle SOMO | LUMO \rangle$')
    plt.legend()
    plt.xlim(3, 2.2)

    plt.figure()
    plt.plot(res_C.coord(), res_C.orbital_energy_gap('Methyl(SOMO)_A', 'Substrate(LUMO)_A'), label='C-addition')
    plt.plot(res_X.coord(), res_X.orbital_energy_gap('Methyl(SOMO)_A', 'Substrate(LUMO)_A'), label='X-addition')
    plt.ylabel(r'$|\epsilon_{SOMO} - \epsilon_{LUMO}|$')
    plt.legend()
    plt.xlim(3, 2.2)

    plt.figure()
    plt.plot(res_C.coord(), res_C.overlap('Methyl(LUMO)_B', 'Substrate(7A_B)_B'), label='C-addition')
    plt.plot(res_X.coord(), res_X.overlap('Methyl(LUMO)_B', 'Substrate(7A_B)_B'), label='X-addition')
    plt.ylabel(r'$\langle SUMO | HOMO \rangle$')
    plt.legend()
    plt.xlim(3, 2.2)

    plt.figure()
    plt.plot(res_C.coord(), res_C.orbital_energy_gap('Methyl(LUMO)_B', 'Substrate(7A_B)_B'), label='C-addition')
    plt.plot(res_X.coord(), res_X.orbital_energy_gap('Methyl(LUMO)_B', 'Substrate(7A_B)_B'), label='X-addition')
    plt.ylabel(r'$|\epsilon_{SUMO} - \epsilon_{HOMO}|$')
    plt.legend()
    plt.xlim(3, 2.2)

    plt.show()



"""Module containing functions for calculating geometrical parameters for molecules"""

from typing import Dict, List, Tuple

import click
from strenum import StrEnum
from tcutility import geometry, molecule


class GeometricParameter(StrEnum):
    """Enum class for geometric parameters"""

    COORDINATE = "Coordinate"
    DISTANCE = "Distance"
    ANGLE = "Angle"
    DIHEDRAL = "Dihedral"
    PYRAMIDAL = "Pyramidal"


geometric_character_to_info_mapping: Dict[GeometricParameter, Tuple[str, str, int]] = {
    GeometricParameter.COORDINATE: ("Coordinate", "Å", 6),
    GeometricParameter.DISTANCE: ("Distance", " Å", 3),
    GeometricParameter.COORDINATE: ("Angle", "°", 2),
    GeometricParameter.DIHEDRAL: ("Dihedral", "°", 2),
    GeometricParameter.PYRAMIDAL: ("Pyramidal", "°", 2),
}


@click.command()
@click.argument("xyzfile", type=click.Path(exists=True))
@click.argument("atom_indices", type=str, nargs=-1)
@click.option("-p", "--pyramidal", is_flag=True, default=False, help="Instead of calculating dihedral angles, calculate pyramidalisation angle.")
def calculate_geometry_parameter(xyzfile: str, atom_indices: List[str], pyramidal: bool) -> None:
    """
    Calculate geometrical parameters for atoms at the provided indices.

    For 1 index this program returns the cartesian coordinate for this atom.
    For 2 indices return bond length between atoms.
    For 3 indices return bond angle, with the second index being the central atom.
    For 4 indices return dihedral angle by calculating the angle between normal vectors described by atoms at indices 1-2-3 and indices 2-3-4.
    If the -p/--pyramidal flag is turned on it calculates 360° - ang1 - ang2 - ang3, where ang1, ang2 and ang3 are the angles described by indices 2-1-3, 3-1-4 and 4-1-2 respectively.
    """
    mol = molecule.load(xyzfile)
    indices = [int(i) - 1 for i in atom_indices]

    assert 1 <= len(indices) <= 4, f"Number of indices must be 1, 2, 3 or 4, not {len(indices)}"

    atoms = "-".join([f"{mol[i+1].symbol}{i+1}" for i in indices])

    param_value = geometry.parameter(mol, *indices, pyramidal=pyramidal)

    if len(indices) == 1:
        param_value = "  ".join([f"{x: .6f}" for x in mol[indices[0] + 1].coords])
        param_type = GeometricParameter.COORDINATE
    elif len(indices) == 2:
        param_type = GeometricParameter.DISTANCE
    elif len(indices) == 3:
        param_type = GeometricParameter.ANGLE
    elif len(indices) == 4 and not pyramidal:
        param_type = GeometricParameter.DIHEDRAL
    elif len(indices) == 4 and pyramidal:
        param_type = GeometricParameter.PYRAMIDAL
    else:
        raise ValueError("Invalid number of indices")

    param_type, unit, precision = geometric_character_to_info_mapping[param_type]
    print(f"{param_type}({atoms}) = {param_value: .{precision}f}{unit}")


if __name__ == "__main__":
    calculate_geometry_parameter()



import click
from tcutility.cli_scripts.cite import generate_citations
from tcutility.cli_scripts.concatenate_irc import concatenate_irc_paths
from tcutility.cli_scripts.geo import calculate_geometry_parameter
from tcutility.cli_scripts.job_script import optimize_geometry
from tcutility.cli_scripts.read import read_results
from tcutility.cli_scripts.resize_figures import resize



@click.group()
def tc():
    """TCutility command line interface."""
    pass


tc.add_command(read_results, name="read")
tc.add_command(optimize_geometry, name="optimize")
tc.add_command(generate_citations, name="cite")
tc.add_command(calculate_geometry_parameter, name="geo")
tc.add_command(concatenate_irc_paths, name="concat-irc")
tc.add_command(resize, name="resize")
